from collections.abc import Hashable, Iterable, Iterator, Sequence
import contextlib
import functools
import itertools
import math
from typing import TYPE_CHECKING, Final, TypeVar

import beartype
import jax
import jaxtyping as jt
import numpy as np

from . import _types, datasets

if TYPE_CHECKING:
    import reax

__all__ = (
    "SequentialSampler",
    "RandomSampler",
    "BatchSampler",
    "IterableSampler",
    "DistributedSampler",
)

_T_co = TypeVar("_T_co", covariant=True)
_IdxT = TypeVar("_IdxT", bound=Hashable)


class SequentialSampler(_types.Sampler[int]):
    """Sequentially sample integers index samples up to a given `length`.

    Equivalent to `range(length)`.
    """

    def __init__(self, length: int) -> None:
        """Init function."""
        if not isinstance(length, int):
            raise TypeError("Length must be an integer")

        self._length: Final[int] = length

    def __iter__(self) -> Iterator[int]:
        """Iter function."""
        return iter(range(self._length))

    def __len__(self) -> int:
        """Len function."""
        return self._length


class RandomSampler(_types.Sampler[int]):
    """Randomly sample integer index up to a given ``length`` with possible ``replacements``."""

    SAMPLE_SIZE = 32  # Used to control the number of samples we generate internally at once

    def __init__(self, length: int, replacements: bool = False, num_samples: int | None = None):
        # Params
        self._length: Final[int] = length
        self._replacements: Final[bool] = replacements
        self._num_samples: Final[bool | None] = num_samples

    @property
    def num_samples(self) -> int:
        """Num samples."""
        if self._num_samples is None:
            return self._length

        return self._num_samples  # Fixed number of samples

    def __len__(self) -> int:
        """Len function."""
        return self.num_samples

    def __iter__(self) -> Iterator[int]:
        """Iter function."""
        total = self._length

        if self._replacements:
            for _ in range(self.num_samples // self.SAMPLE_SIZE):
                yield from np.random.randint(0, high=total, size=(self.SAMPLE_SIZE,)).tolist()
            yield from np.random.randint(
                0, high=total, size=(self.num_samples % self.SAMPLE_SIZE,)
            ).tolist()
        else:
            for _ in range(self.num_samples // total):
                yield from np.random.permutation(total).tolist()
            yield from np.random.permutation(total).tolist()[: self.num_samples % total]


class BatchSampler(_types.Sampler[list[_IdxT]]):
    r"""Sample batches of indexes from a given sample."""

    def __init__(self, sampler: _types.Sampler[_IdxT], batch_size: int, drop_last: bool) -> None:
        # Params
        self._batch_size: Final[int] = batch_size
        self._drop_last: Final[bool] = drop_last

        # State
        self._sampler = sampler

    def __iter__(self) -> Iterator[list[_IdxT]]:
        """Iter function."""
        if self._drop_last:
            sampler_iter = iter(self._sampler)
            while True:
                try:
                    batch = [next(sampler_iter) for _ in range(self._batch_size)]
                    yield batch
                except StopIteration:
                    break
        else:
            batch = [0] * self._batch_size
            idx_in_batch = 0
            for idx in self._sampler:
                batch[idx_in_batch] = idx
                idx_in_batch += 1
                if idx_in_batch == self._batch_size:
                    yield batch
                    idx_in_batch = 0
                    batch = [0] * self._batch_size
            if idx_in_batch > 0:
                yield batch[:idx_in_batch]

    def __len__(self) -> int:
        """Len function."""
        if self._drop_last:
            return len(self._sampler) // self._batch_size

        return (len(self._sampler) + self._batch_size - 1) // self._batch_size


class IterableSampler(_types.Sampler[list[None]]):
    def __iter__(self) -> Iterator[list[None]]:
        """Iter function."""
        yield from itertools.repeat([None])


class DistributedSampler(_types.Sampler[_IdxT]):
    """A sampler that distributes data across multiple processes, ensuring each process
    receives a unique and non-overlapping subset of the dataset.

    This sampler is designed for use with JAX's distributed training capabilities.
    It handles shuffling, dropping the last incomplete batch (if specified),
    and partitioning the dataset based on the process rank and number of replicas.
    """

    @jt.jaxtyped(typechecker=beartype.beartype)
    def __init__(
        self,
        dataset: "reax.data.Dataset",
        num_replicas: int | None = None,
        process_index: int | None = None,
        shuffle: bool = True,
        seed: int = 0,
        drop_last: bool = False,
    ) -> None:
        """Initializes the DistributedSampler.

        Args:
            dataset: The dataset to sample from.
            num_replicas: The total number of replicas (processes).
                Defaults to jax.process_count().
            process_index: The index of the current process. Defaults to jax.process_index().
            shuffle: Whether to shuffle the data.
            seed: The seed for shuffling.
            drop_last: Whether to drop the last incomplete batch.
        """
        if num_replicas == 0:
            raise ValueError("Number of replicas cannot be 0.")

        # Params
        self._num_replicas: Final[int] = self._init_num_replicas(num_replicas)
        self._process_index: Final[int] = self._init_process_index(process_index)
        if self._process_index >= self._num_replicas:
            raise ValueError(
                f"Process index ({self._process_index}) must be less than the number of replicas "
                f"({self._num_replicas})."
            )

        self._shuffle: Final[bool] = shuffle
        self._drop_last: Final[bool] = drop_last
        self._len_dataset = datasets.len_dataset(dataset)
        self._num_samples: Final[float] = self._init_num_samples(
            self._len_dataset, drop_last, self._num_replicas
        )
        self._total_size: Final[int] = self._num_samples * self._num_replicas
        self._seed: Final[int] = seed

        # State
        self._epoch: int = 0

    @staticmethod
    def _init_num_samples(len_dataset: int | float, drop_last: bool, num_replicas: int) -> float:
        """Calculates the number of samples each process should receive.

        If the dataset length is evenly divisible by the number of replicas,
        no data is dropped. Otherwise, the dataset is split to the nearest
        available length that is evenly divisible.

        Args:
            len_dataset: The total size of the dataset.
            drop_last: Whether to drop the last incomplete batch.
            num_replicas: The number of replicas.

        Returns:
            The number of samples each process should receive.
        """
        # If the dataset length is evenly divisible by # of replicas, then there
        # is no need to drop any data, since the dataset will be split equally.
        if drop_last and len_dataset % num_replicas != 0:
            # Split to nearest available length that is evenly divisible.
            # This is to ensure each rank receives the same amount of data when
            # using this Sampler.
            return math.ceil((len_dataset - num_replicas) / num_replicas)

        return math.ceil(len_dataset / num_replicas)

    @staticmethod
    def _init_num_replicas(num_replicas: int | None) -> int:
        return num_replicas if num_replicas is not None else jax.process_count()

    @staticmethod
    def _init_process_index(process_index: int | None) -> int:
        return process_index if process_index is not None else jax.process_index()

    def __iter__(self) -> Iterator[_IdxT]:
        """Returns an iterator over the sampled indices.

        The indices are shuffled (if specified) and partitioned across processes.
        If `drop_last` is True, the last incomplete batch is removed.  Otherwise,
        the dataset is padded to ensure each process receives the same number of samples.

        Returns:
            An iterator over the sampled indices.
        """
        if self._shuffle:
            # deterministically shuffle based on epoch and seed
            key = jax.random.key(self._seed + self._epoch)
            indices = jax.random.permutation(key, self._len_dataset).tolist()
        else:
            indices = list(range(self._len_dataset))

        if self._drop_last:
            # remove the last set of indices
            indices = indices[: self._total_size]
        else:
            # pad up to evenly divisible
            padding_size = self._total_size - len(indices)
            if padding_size <= len(indices):
                indices += indices[:padding_size]
            else:
                indices += (indices * math.ceil(padding_size / len(indices)))[:padding_size]

        assert len(indices) == self.total_size

        # subsample
        indices = indices[self._process_index : self._total_size : self._num_replicas]
        assert len(indices) == self._num_samples

        return iter(indices)

    def __len__(self) -> int:
        """Returns the number of samples in the sampler.

        Returns:
            The number of samples.
        """
        return self._num_samples

    @property
    def total_size(self) -> int:
        return self._total_size

    def set_epoch(self, epoch: int) -> None:
        """Sets the epoch for shuffling.

        This allows for different shuffles in each epoch.

        Args:
            epoch: The epoch number.
        """
        self._epoch = epoch


def create_sampler(
    dataset: _types.Dataset[_T_co],
    batch_size: int | None = None,
    replacements: bool = False,
    shuffle: bool = False,
    sampler: "reax.data.Sampler[_T_co]" = None,
    drop_last: bool = False,
) -> "reax.data.Sampler[_T_co]":
    """Create sampler."""
    if sampler is None:
        # Need this special case because jax arrays are not really subclasses of jax.Array and won't
        # be matched by singledispatch
        if isinstance(dataset, jax.Array):
            sampler = create_sequence_sampler(dataset, replacements=replacements, shuffle=shuffle)
        else:
            sampler = _create_sampler(dataset, replacements=replacements, shuffle=shuffle)

    if batch_size is not None:
        sampler = BatchSampler(sampler, batch_size, drop_last)

    return sampler


@functools.singledispatch
def _create_sampler(
    dataset: _types.Dataset[_T_co], replacements: bool = False, shuffle: bool = False
) -> "reax.data.Sampler[_T_co]":
    """Create sampler."""
    raise TypeError(f"Unsupported type {type(dataset).__name__}")


with contextlib.suppress(ImportError):
    import torch.utils.data

    @_create_sampler.register(torch.utils.data.IterableDataset)
    def create_torch_iterable_dataset_sampler(
        dataset: torch.utils.data.IterableDataset[_T_co],
        replacements: bool = False,
        shuffle: bool = False,
    ):
        """Create torch iterable dataset sampler."""
        return create_iterable_sampler(dataset, replacements=replacements, shuffle=shuffle)

    @_create_sampler.register(torch.utils.data.Dataset)
    def create_torch_dataset_sampler(
        dataset: torch.utils.data.Dataset[_T_co], replacements: bool = False, shuffle: bool = False
    ):
        """Create torch dataset sampler."""
        return create_sequence_sampler(dataset, replacements=replacements, shuffle=shuffle)


@_create_sampler.register(Sequence)
@_create_sampler.register(np.ndarray)
def create_sequence_sampler(
    dataset, replacements: bool = False, shuffle: bool = False
) -> "reax.data.Sampler[_T_co]":
    """Create sequence sampler."""
    if shuffle:
        return RandomSampler(len(dataset), replacements=replacements)

    return SequentialSampler(len(dataset))


@_create_sampler.register(Iterable)
def create_iterable_sampler(
    dataset: Iterable[_T_co], replacements: bool = False, shuffle: bool = False
) -> _types.Sampler[None] | _types.Sampler[list[None]]:
    """Create iterable sampler."""
    if shuffle:
        raise ValueError(
            f"``shuffle=True`` is not supported with dataset type {type(dataset).__name__} which "
            f"does not support random access"
        )
    if replacements:
        raise ValueError(
            f"``replacements=True`` is not supported with dataset type {type(dataset).__name__} "
            f"which does not support random access"
        )

    return IterableSampler()


def create_batch_sampler(
    dataset: Sequence[_T_co], replacements: bool = False, shuffle: bool = False
) -> BatchSampler[int]:
    """Create batch sampler."""
    if shuffle:
        return RandomSampler(len(dataset), replacements=replacements)

    return SequentialSampler(len(dataset))



import functools
from typing import TYPE_CHECKING, Any

import equinox
import jax
import jax.numpy as jnp
import jaxtyping as jt
import optax

if TYPE_CHECKING:
    import reax

__all__ = ("Optimizer",)


class Optimizer(equinox.Module):
    optimizer: optax.GradientTransformation
    state: optax.OptState
    update_count: (
        int | jt.Int[jax.Array, ""]
    )  # Total number of times this optimizer was used to parameters

    @classmethod
    def from_params(cls, opt: optax.GradientTransformation, params: optax.Params) -> "Optimizer":
        """From params."""
        return cls(opt, opt.init(params))

    def __init__(
        self,
        opt: optax.GradientTransformation,
        state: optax.OptState,
        count: int | jt.Int[jax.Array, ""] = 0,
    ):
        """Init function."""
        self.optimizer = opt
        self.state = state
        self.update_count = count

    @equinox.filter_jit
    def update(
        self, params: optax.Params, grad: jt.PyTree, value: float | None = None
    ) -> tuple[Any, "Optimizer"]:
        """Return an updated version of the passed parameters using the passed gradients.

        Returns:
            tuple[Any, "Optimizer"]: A tuple of the updated parameters
            and an instance of this optimizer updated with the new
            state.
        """
        updates, new_state = self.optimizer.update(
            grad,
            self.state,
            params=params,
            extra_args={"value": value} if value is not None else {},
        )

        params = optax.apply_updates(params, updates)

        count = getattr(new_state, "gradient_step", self.update_count + 1)
        return params, type(self)(self.optimizer, new_state, count=count)

    def update_module(
        self, module: "reax.Module", grad: jt.PyTree, value: float | None = None
    ) -> "Optimizer":
        """Perform an inplace update of the module parameters given the passed gradients.

        Returns:
            "Optimizer": An instance of this optimizer updated with the
            new state.
        """
        new_params, new_state = _update(
            self.optimizer, self.state, grad, module.parameters(), value=value
        )
        module.set_parameters(new_params)
        count = getattr(new_state, "gradient_step", self.update_count + 1)
        return type(self)(self.optimizer, new_state, count=count)


class DistributedOptimizer(Optimizer):
    def __init__(
        self,
        opt: optax.GradientTransformation,
        state: optax.OptState,
        engine: "reax.Engine",
        count: int | jt.Int[jax.Array, ""] = 0,
    ):
        super().__init__(opt, state, count)
        self._engine = engine

    @equinox.filter_jit
    def update(
        self, params: optax.Params, grad: jt.PyTree, value: float | None = None
    ) -> tuple[Any, "Optimizer"]:
        """Return an updated version of the passed parameters using the passed gradients.

        Returns:
            tuple[Any, "Optimizer"]: A tuple of the updated parameters
            and an instance of this optimizer updated with the new
            state.
        """
        # Average the gradients across processes
        grad = self._engine.all_reduce(grad, reduce_op="mean")
        return super().update(params, grad, value=value)

    def update_module(
        self, module: "reax.Module", grad: jt.PyTree, value: float | None = None
    ) -> "Optimizer":
        """Perform an inplace update of the module parameters given the passed gradients.

        Returns:
            "Optimizer": An instance of this optimizer updated with the
            new state.
        """
        new_params, new_state = _update(
            self.optimizer, self.state, grad, module.parameters(), value=value
        )
        module.set_parameters(new_params)
        count = getattr(new_state, "gradient_step", self.update_count + 1)
        return type(self)(self.optimizer, new_state, engine=self._engine, count=count)


@functools.partial(jax.jit, static_argnames=("optimizer",), donate_argnames="params")
def _update(
    optimizer: optax.GradientTransformation, state, grad: dict, params, value: float | None = None
):
    """Jax jitted function that performs an optimizer update based on the passed gradients and
    parameters.
    """
    updates, new_state = optimizer.update(grad, state, params=params, value=value)
    params = optax.apply_updates(params, updates)

    return params, new_state


mock_optimizer = optax.GradientTransformation(
    init=lambda params: {},
    update=lambda grad, opt_state, *_, **__: (jax.tree.map(jnp.zeros_like, grad), opt_state),
)



from typing import TYPE_CHECKING, Generic, Optional, TypeVar

from .. import exceptions

if TYPE_CHECKING:
    import reax

__all__ = ("DataSource",)

_T_co = TypeVar("_T_co", covariant=True)


class DataSource(Generic[_T_co]):
    """Interface for classes that provide datasets."""

    def __init__(self) -> None:
        """Attributes:
        prepare_data_per_node:
            If True, each LOCAL_RANK=0 will call prepare data.
            Otherwise, only NODE_RANK=0, LOCAL_RANK=0 will prepare data.
        allow_zero_length_dataloader_with_multiple_devices:
            If True, dataloader with zero length within local rank is allowed.
            Default value is False.
        """
        super().__init__()
        self.prepare_data_per_node: bool = True
        self.allow_zero_length_dataloader_with_multiple_devices: bool = False

    # Region Events

    def prepare_data(self) -> None:
        """Use this to download and prepare data. Downloading and saving data with multiple
        processes (distributed settings) will result in corrupted data. REAX ensures this
        method is called only within a single process, so you can safely add your downloading logic
        within.

        Warning:
            DO NOT set state to the model (use ``setup`` instead)
            since this is NOT called on every device

        Example::

            def prepare_data(self):
                # good
                download_data()
                tokenize()
                etc()

                # bad
                self.split = data_split
                self.some_state = some_other_state()

        In a distributed environment, ``prepare_data`` can be called in two ways
        (using :ref:`prepare_data_per_node<common/lightning_module:prepare_data_per_node>`)

        1. Once per node. This is the default and is only called on LOCAL_RANK=0.
        2. Once in total. Only called on GLOBAL_RANK=0.

        Example::

            # DEFAULT
            # called once per node on LOCAL_RANK=0 of that node
            class ReDataModule(ReaxDataModule):
                def __init__(self):
                    super().__init__()
                    self.prepare_data_per_node = True

            # call on GLOBAL_RANK=0 (great for shared file systems)
            class ReDataModule(ReaxDataModule):
                def __init__(self):
                    super().__init__()
                    self.prepare_data_per_node = False

        This is called before requesting the dataloaders:

        .. code-block:: python

            model.prepare_data()
            initialize_distributed()
            model.setup(stage)
            model.train_dataloader()
            model.val_dataloader()
            model.test_dataloader()
            model.predict_dataloader()

        """

    def setup(self, stage: "reax.Stage", /) -> None:
        """Called at the beginning of a trainer stage. This is a good hook when you need to
        build models dynamically or adjust something about them. This hook is called on every
        process when using DDP.

        Example::

            class LitModel(...):
                def __init__(self):
                    self.l1 = None

                def prepare_data(self):
                    download_data()
                    tokenize()

                    # don't do this
                    self.something = else

                def setup(self, stage):
                    data = load_data(...)
                    self.l1 = nn.Linear(28, data.num_classes)

        """

    def teardown(self, stage: "reax.Stage", /) -> None:
        """Called at the end of a trainer stage"""

    def on_exception(self, exception: BaseException, /) -> None:
        """Called when the stage execution is interrupted by an exception."""

    # endregion

    def train_dataloader(self) -> "reax.DataLoader[_T_co]":
        """An iterable or collection of iterables specifying training samples.

        For more information about multiple dataloaders, see :ref:`section <multiple-dataloaders>`.

        The dataloader you return will not be reloaded unless you set
        :paramref:`~reax.Trainer.fit(reload_dataloaders_every_n_epochs=k)` to a positive integer.

        For data processing use the following pattern:

            - download in :meth:`prepare_data`
            - process and split in :meth:`setup`

        However, the above are only necessary for distributed processing.

        Warning:
            do not assign state in prepare_data

        - :meth:`~reax.Trainer.fit`
        - :meth:`prepare_data`
        - :meth:`setup`

        Note:
            REAX tries to add the correct sampler for distributed and arbitrary hardware.
            There is no need to set it yourself.

        """
        raise exceptions.MisconfigurationException(
            "`train_dataloader` must be implemented to be used with the REAX Trainer"
        )

    def val_dataloader(self) -> Optional["reax.DataLoader[_T_co]"]:
        r"""An iterable or collection of iterables specifying validation samples.

        For more information about multiple dataloaders, see :ref:`section <multiple-dataloaders>`.

        The dataloader you return will not be reloaded unless you set
        :paramref:`~lightning.pytorch.trainer.trainer.Trainer.reload_dataloaders_every_n_epochs` to
        a positive integer.

        It's recommended that all data downloads and preparation happen in :meth:`prepare_data`.

        - :meth:`~reax.Trainer.fit`
        - :meth:`~reax.Trainer.validate`
        - :meth:`prepare_data`
        - :meth:`setup`

        Note:
            REAX tries to add the correct sampler for distributed and arbitrary hardware
            There is no need to set it yourself.

        Note:
            If you don't need a validation dataset and a :meth:`validation_step`, you don't need to
            implement this method.

        """

    def test_dataloader(self) -> "reax.DataLoader[_T_co]":
        r"""An iterable or collection of iterables specifying test samples.

        For more information about multiple dataloaders, this :ref:`section <multiple-dataloaders>`.

        For data processing use the following pattern:

            - download in :meth:`prepare_data`
            - process and split in :meth:`setup`

        However, the above are only necessary for distributed processing.

        Warning:
            do not assign state in prepare_data

        - :meth:`~lightning.pytorch.trainer.trainer.Trainer.test`
        - :meth:`prepare_data`
        - :meth:`setup`

        Note:
            REAX tries to add the correct sampler for distributed and arbitrary hardware.
            There is no need to set it yourself.

        Note:
            If you don't need a test dataset and a :meth:`test_step`, you don't need to implement
            this method.

        """
        raise exceptions.MisconfigurationException(
            "`test_dataloader` must be implemented to be used with the REAX Trainer"
        )

    def predict_dataloader(self) -> "reax.DataLoader[_T_co]":
        r"""An iterable or collection of iterables specifying prediction samples.

        For more information about multiple dataloaders, see :ref:`section <multiple-dataloaders>`.

        It's recommended that all data downloads and preparation happen in :meth:`prepare_data`.

        - :meth:`~reax.Trainer.predict`
        - :meth:`prepare_data`
        - :meth:`setup`

        Note:
            REAX tries to add the correct sampler for distributed and arbitrary hardware
            There is no need to set it yourself.

        Return:
            A :class:`reax.DataLoader` or a sequence of them specifying prediction samples.

        """
        raise exceptions.MisconfigurationException(
            "`predict_dataloader` must be implemented to be used with the REAX Trainer"
        )



from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import reax


class ModuleHooks:
    # pylint: disable=too-many-public-methods

    def on_fit_start(self, stage: "reax.stages.Fit", /) -> None:
        """A fitting stage is about to begin."""

    def on_fit_end(self, stage: "reax.stages.Fit", /) -> None:
        """A fitting stage is about to end."""

    # region Train

    def on_train_start(self, stage: "reax.stages.Train", /) -> None:
        """Training is about to start."""

    def on_train_epoch_start(self, stage: "reax.stages.Train", /) -> None:
        """A training epoch is about to begin."""

    def on_train_batch_start(
        self, stage: "reax.stages.Train", batch: Any, batch_idx: int, /
    ) -> None:
        """The training stage if about to process a batch."""

    def on_train_batch_end(
        self, stage: "reax.stages.Train", outputs: Any, batch: Any, batch_idx: int, /
    ) -> None:
        """The training stage has just finished processing a batch."""

    def on_train_epoch_end(self, stage: "reax.stages.Train", /) -> None:
        """A training epoch is about to end."""

    def on_train_end(self, stage: "reax.stages.Train", /) -> None:
        """Training is about to end."""

    # endregion

    # region Validation

    def on_validation_start(self, stage: "reax.stages.Validate", /) -> None:
        """Validation is starting."""

    def on_validation_epoch_start(self, stage: "reax.stages.Validate", /) -> None:
        """A validation epoch is about to begin."""

    def on_validation_batch_start(
        self, stage: "reax.stages.Validate", batch: Any, batch_idx: int
    ) -> None:
        """The validation stage if about to process a batch."""

    def on_validation_batch_end(
        self, stage: "reax.stages.Validate", outputs: Any, batch: Any, batch_idx: int, /
    ) -> None:
        """The validation stage has just finished processing a batch."""

    def on_validation_epoch_end(self, stage: "reax.stages.Validate", /) -> None:
        """A validation epoch is about to end."""

    def on_validation_end(self, stage: "reax.stages.Validate", /) -> None:
        """Validation has ended."""

    # endregion

    # region Test

    def on_test_start(self, stage: "reax.stages.Test", /) -> None:
        """The test stage is starting"""

    def on_test_epoch_start(self, stage: "reax.stages.Test", /) -> None:
        """A test epoch is about to begin."""

    def on_test_batch_start(self, stage: "reax.stages.Test", batch: Any, batch_idx: int, /) -> None:
        """The test stage if about to process a batch."""

    def on_test_batch_end(
        self, stage: "reax.stages.Test", outputs: Any, batch: Any, batch_idx: int, /
    ) -> None:
        """The test stage has just finished processing a batch."""

    def on_test_epoch_end(self, stage: "reax.stages.Test", /) -> None:
        """A test epoch is about to end."""

    def on_test_end(self, stage: "reax.stages.Test", /) -> None:
        """The test stage has finished"""

    # endregion

    # region Predict

    def on_predict_start(self, stage: "reax.stages.Predict", /) -> None:
        """Prediction is starting."""

    def on_predict_epoch_start(self, stage: "reax.stages.Predict", /) -> None:
        """A predict epoch is about to begin."""

    def on_predict_batch_start(
        self, stage: "reax.stages.Predict", batch: Any, batch_idx: int, /
    ) -> None:
        """The test stage if about to process a batch."""

    def on_predict_batch_end(
        self, stage: "reax.stages.Predict", outputs: Any, batch: Any, batch_idx: int, /
    ) -> None:
        """The predict stage has just finished processing a batch."""

    def on_predict_epoch_end(self, stage: "reax.stages.Train", /) -> None:
        """A predict epoch is about to end."""

    def on_predict_end(self, stage: "reax.stages.Predict", /) -> None:
        """Prediction is starting."""

    # endregion

    def on_before_optimizer_step(
        self, optimizer: "reax.Optimizer", grad: dict[str, Any], /
    ) -> None:
        """Called before ``optimizer.step()``.

        If using gradient accumulation, the hook is called once the gradients have been accumulated.
        See: :paramref:`~reax.Trainer.accumulate_grad_batches`.

        If clipping gradients, the gradients will not have been clipped yet.

        Args:
            optimizer: Current optimizer being used.
            grad: The gradients dictionary from JAX
        """

    # region Generic stage messages

    def on_stage_start(self, stage: "reax.stages.Stage", /) -> None:
        """A stage is about to start."""

    def on_stage_iter_starting(self, stage: "reax.Stage", step: int, /) -> None:
        """A stage is about to start an interation."""

    def on_stage_iter_ending(self, stage: "reax.Stage", step: int, outputs: Any, /):
        """The stage just finished processing an iteration."""

    def on_stage_end(self, stage: "reax.Stage", /) -> None:
        """The stage is ending."""

    # endregion



import math

__all__ = ("convert_inf",)


def convert_inf(x: int | float | None) -> int | float | None:
    """For code that doesn't support inf/nan values such as tqdm -> convert these to `None`."""
    if x is None or math.isinf(x) or math.isnan(x):
        return None
    return x



# Copyright (C) 2024  Martin Uhrin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Most of this file is covered by the following license.  To find what has been modified you
# can perform a diff with the file at:
# https://github.com/Lightning-AI/pytorch-lightning/blob/0324a20f00235c7a10a235a44326811ba42b6ae4/src/lightning/pytorch/loggers/tensorboard.py
#
# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""TensorBoard Logger
------------------
"""

import argparse
from collections.abc import Callable, Mapping
import os
from typing import Any, Final

import fsspec
import jax.typing
import numpy as np
import tensorboardX
from typing_extensions import override

try:
    import omegaconf
except ImportError:
    omegaconf = None

from reax.lightning import rank_zero

from . import _utils, logger
from .. import types

__all__ = ("TensorBoardLogger",)


class TensorBoardLogger(logger.WithDdp["tensorboardX.SummaryWriter"], logger.Logger):
    r"""Log to local or remote file system in
    `TensorBoard <https://www.tensorflow.org/tensorboard>`_ format.

    Implemented using :class:`~tensorboardX.SummaryWriter`. Logs are saved to
    ``os.path.join(save_dir, name, version)``. This is the default logger in Lightning, it comes
    preinstalled.

    This logger supports logging to remote filesystems via ``fsspec``. Make sure you have it
    installed and you don't have tensorflow (otherwise it will use tf.io.gfile instead of fsspec).

    Example:

    .. testcode::
        :skipif: not _TENSORBOARD_AVAILABLE or not _TENSORBOARDX_AVAILABLE

        from reax import Trainer
        from reax.loggers import TensorBoardLogger

        logger = TensorBoardLogger("tb_logs", name="my_model")
        trainer = Trainer(logger=logger)
    """

    LOGGER_JOIN_CHAR: Final[str] = "-"
    NAME_HPARAMS_FILE: Final[str] = "hparams.yaml"

    def __init__(
        self,
        log_dir: types.Path,
        name: str | None = "reax_logs",
        *,
        version: int | str | None = None,
        log_graph: bool = False,
        default_hp_metric: bool = True,
        prefix: str = "",
        sub_dir: types.Path | None = None,
        **kwargs: Any,
    ):
        """Init function.

        Args:
            **kwargs (Any): Additional arguments used by
                :class:`tensorboardX.SummaryWriter` can be passed as
                keyword arguments in this logger. To automatically flush
                to disk, `max_queue` sets the sizeof the queue for
                pending logs before flushing. `flush_secs` determines
                how many seconds elapses before flushing.
            log_dir (typing.Path)
            save_dir: Save directory.
            name (str | None, optional): Experiment name.  If it is the
                empty string then no per-experiment subdirectory is
                used, defaults to "reax_logs".
            version (int | str | None, optional): Experiment version. If
                version is not specified the logger inspects the save
                directory for existing versions, then automatically
                assigns the next available version. If it is a string
                then it is used as the run-specific subdirectory name,
                otherwise ``'version_${version}'`` is used, defaults to
                None.
            log_graph (bool, optional): Adds the computational graph to
                tensorboard. This requires that the user has defined the
                `self.example_input_array` attribute in their model,
                defaults to False.
            default_hp_metric (bool, optional): Enables a placeholder
                metric with key `hp_metric` when `log_hyperparams` is
                called without a metric (otherwise calls to
                log_hyperparams without a metric are ignored), defaults
                to True.
            prefix (str, optional): A string to put at the beginning of
                metric keys, defaults to "".
            sub_dir (typing.Path | None, optional): Sub-directory to
                group TensorBoard logs. If a sub_dir argument is passed
                then logs are saved in
                ``/save_dir/name/version/sub_dir/``. logs are saved in
                ``/save_dir/name/version/``, defaults to None.
        """
        super().__init__()
        self._root_dir = os.fspath(log_dir)
        self._name = name or ""
        self._version = version
        self._sub_dir = None if sub_dir is None else os.fspath(sub_dir)

        self._default_hp_metric = default_hp_metric
        self._prefix = prefix
        self._fs: fsspec.AbstractFileSystem = fsspec.url_to_fs(log_dir)[0]

        self._exp: "tensorboardX.SummaryWriter | None" = None
        self._kwargs = kwargs
        self._should_log_graph = log_graph
        self.hparams: dict[str, Any] | argparse.Namespace = {}

    @property
    @override
    def name(self) -> str:
        """Name function."""
        return self._name

    @property
    @override
    def version(self) -> int | str:
        """Version function."""
        if self._version is None:
            self._version = self._get_next_version()
        return self._version

    @property
    @override
    def root_dir(self) -> str:
        """Parent directory for all tensorboard checkpoint subdirectories.

        If the experiment name parameter is an empty string, no experiment subdirectory is used and
        the checkpoint will be saved in "save_dir/version"
        """
        return os.path.join(self._root_dir, self.name)

    @property
    @override
    def log_dir(self) -> str:
        """The directory for this run's tensorboard checkpoint.

        By default, it is named ``'version_${self.version}'`` but it can be overridden by passing a
        string value for the constructor's version parameter instead of ``None`` or an int.
        """
        # create a pseudo standard path ala test-tube
        version = self.version if isinstance(self.version, str) else f"version_{self.version}"
        log_dir = os.path.join(self.root_dir, version)
        if isinstance(self.sub_dir, str):
            log_dir = os.path.join(log_dir, self.sub_dir)

        log_dir = os.path.expandvars(log_dir)
        log_dir = os.path.expanduser(log_dir)
        return log_dir

    @property
    def sub_dir(self) -> str | None:
        """Sub dir."""
        return self._sub_dir

    @property
    @override
    def save_dir(self) -> str:
        """Gets the save directory where the TensorBoard experiments are saved.

        Returns:
            The local path to the save directory where the TensorBoard experiments are saved.
        """
        return self._root_dir

    @property
    def _experiment(self) -> "tensorboardX.SummaryWriter | None":
        """Get the tensorboard object."""
        if self._exp is not None:
            return self._exp

        assert rank_zero.rank_zero_only.rank == 0, "tried to init log dirs in non process_index=0"

        if self.root_dir:
            self._fs.makedirs(self.root_dir, exist_ok=True)

        self._exp = tensorboardX.SummaryWriter(log_dir=self.log_dir, **self._kwargs)
        return self._exp

    @override
    def _log_metrics(
        self, metrics: Mapping[str, jax.typing.ArrayLike], step: int | None = None
    ) -> None:
        """Log metrics."""
        metrics = _utils.add_prefix(metrics, self._prefix, self.LOGGER_JOIN_CHAR)

        for key, val in metrics.items():
            if isinstance(val, (jax.Array, np.ndarray)):
                val = val.item()

            if isinstance(val, dict):
                self.experiment.add_scalars(key, val, step)
            else:
                try:
                    self.experiment.add_scalar(key, val, step)
                except Exception as ex:
                    raise ValueError(
                        f"Logging of type `{type(val).__name__}` is not supported. "
                        f"Use `dict`, scalar or array"
                    ) from ex

    @override
    def _log_hyperparams(
        # pylint: disable=arguments-differ
        self,
        params: dict[str, Any] | argparse.Namespace,
        metrics: dict[str, Any] | None = None,
    ) -> None:
        """Record hyperparameters.

        TensorBoard logs with and without saved hyperparameters are incompatible, the
        hyperparameters are then not displayed in the TensorBoard. Please delete or move the
        previously saved logs to display the new ones with hyperparameters.

        Args:
            params (dict[str, Any] | argparse.Namespace): A dictionary-
                like container with the hyperparameters.
            metrics (dict[str, Any] | None, optional): Dictionary with
                metric names as keys and measured quantities as values,
                defaults to None.
            step: Optional global step number for the logged metrics.
        """
        params = _utils.convert_params(params)

        if omegaconf is not None and isinstance(params, omegaconf.Container):
            # store params to output
            self.hparams = omegaconf.OmegaConf.merge(self.hparams, params)
        else:
            self.hparams.update(params)

        # format params into the suitable for tensorboard
        params = _utils.flatten_dict(params)
        params = self._sanitize_params(params)

        if metrics is None:
            if self._default_hp_metric:
                metrics = {"hp_metric": -1}
        elif not isinstance(metrics, dict):
            metrics = {"hp_metric": metrics}

        if metrics:
            self.log_metrics(metrics, 0)

            exp, ssi, sei = tensorboardX.summary.hparams(params, metrics)
            writer = self.experiment._get_file_writer()  # pylint: disable=protected-access
            writer.add_summary(exp)
            writer.add_summary(ssi)
            writer.add_summary(sei)

    @override
    def _log_graph(
        # pylint: disable=arguments-differ
        self,
        model: Callable,
        *inputs,
    ) -> None:
        """Log graph."""
        if not self._should_log_graph:
            return

        input_array = model.example_input_array if not inputs else inputs

        if input_array is None:
            rank_zero.rank_zero_warn(
                "Could not log computational graph to TensorBoard: The `model.example_input_array` "
                "attribute is not set or `input_array` was not given."
            )
        else:
            # TODO: Complete this
            # comp = jax.jit(model).lower(inputs).compiler_ir("hlo")
            # self.experiment.file_writer.add_graph
            pass

    @override
    def _save(self) -> None:
        """Save function."""
        self.experiment.flush()

    @override
    def _finalize(self, status: str) -> None:
        """Finalize function."""
        if status == "success":
            # saving hparams happens independent of experiment manager
            self.save()

        if self._exp is not None:
            self.experiment.flush()
            self.experiment.close()

    def _get_next_version(self) -> int:
        """Get next version."""
        root_dir = self.root_dir

        try:
            listdir_info = self._fs.listdir(root_dir)
        except OSError:
            return 0

        existing_versions = []
        for info in listdir_info:
            name = info["name"]
            basename = os.path.basename(name)
            if os.path.isdir(name) and basename.startswith("version_"):
                dir_ver = basename.split("_")[1].replace("/", "")
                if dir_ver.isdigit():
                    existing_versions.append(int(dir_ver))
        if len(existing_versions) == 0:
            return 0

        return max(existing_versions) + 1

    @staticmethod
    def _sanitize_params(params: dict[str, Any]) -> dict[str, Any]:
        """Sanitize params."""
        params = _utils.sanitize_params(params)
        # logging of arrays with dimension > 1 is not supported, sanitize as string
        return {
            key: str(val) if hasattr(val, "ndim") and val.ndim > 1 else val
            for key, val in params.items()
        }



from typing import TYPE_CHECKING, Any
import weakref

import beartype
import jax
import jaxtyping as jt
from typing_extensions import override

from . import stages

if TYPE_CHECKING:
    import reax

__all__ = ("Predict",)


class Predict(stages.EpochStage):
    @jt.jaxtyped(typechecker=beartype.beartype)
    def __init__(
        self,
        module: "reax.Module",
        datamanager: "reax.data.DataSourceManager",
        engine: "reax.Engine",
        *,
        fast_dev_run: bool | int = False,
        limit_batches: int | None = None,
        keep_predictions=True,
    ):
        """Init function."""
        super().__init__(
            "predict",
            module,
            datamanager,
            engine,
            fast_dev_run=fast_dev_run,
            limit_batches=limit_batches,
        )
        self._keep_predictions = keep_predictions
        self._all_outputs: list[Any] | list[list[Any]] | None = []

    @property
    def predictions(self) -> list[Any] | list[list[Any]] | None:
        return self._all_outputs

    @override
    def _on_starting(self):
        super()._on_starting()
        self._module.on_predict_start(weakref.proxy(self))

    @override
    def _on_epoch_start(self):
        super()._on_epoch_start()
        self._module.on_predict_epoch_start(weakref.proxy(self))

    @override
    def _on_iteration_starting(self):
        super()._on_iteration_starting()
        self._module.on_predict_batch_start(self, self.batch, self.batch_idx)

    @override
    def _step(self) -> "reax.stages.MetricResults":
        """Step function."""
        return self._module.predict_step(self.batch, self._iter)

    @override
    def _on_iteration_finishing(self, outputs: Any, /):
        """On iteration finishing."""
        super()._on_iteration_finishing(outputs)
        if self._keep_predictions:
            cpu = jax.devices("cpu")[0]
            self._all_outputs.append(jax.device_put(outputs, cpu))
        self._module.on_predict_batch_end(self, outputs, self.batch, self.batch_idx)

    @override
    def _on_epoch_end(self) -> None:
        super()._on_epoch_end()
        self._module.on_predict_epoch_end(weakref.proxy(self))

    @override
    def _on_stopped(self) -> None:
        super()._on_stopped()
        self._module.on_predict_end(weakref.proxy(self))



# Copyright (C) 2024  Martin Uhrin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Most of this file is covered by the following license.  To find what has been modified you
# can perform a diff with the file at:
# https://github.com/Lightning-AI/pytorch-lightning/blob/9177ec09caadcf88859e1f1e3e10a18e8832069a/tests/tests_pytorch/helpers/simple_models.py
#
# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses

import beartype
import jax.numpy as jnp
import jax.typing
import jaxtyping as jt
from typing_extensions import override

from . import _metric, jm

__all__ = ("Accuracy",)


class Accuracy(_metric.Metric):
    r"""Computes Accuracy, ported from [torchmetrics](https://github.com/PytorchLightning/metrics).

    .. math::
        \text{Accuracy} = \frac{1}{N}\sum_i^N 1(y_i = \hat{y}_i)

    Where :math:`y` is a tensor of target values, and :math:`\hat{y}` is a
    tensor of predictions.

    For multi-class and multidimensional multi-class data with probability or logits predictions,
    the parameter `top_k` generalizes this metric to a Top-K accuracy metric: for each sample the
    top-K highest probability or logit score items are considered to find the correct label.

    For multi-label and multidimensional multi-class inputs, this metric computes the "glob"
    accuracy by default, which counts all target or sub-samples separately. This can be
    changed to subset accuracy (which requires all target or sub-samples in the sample to
    be correctly predicted) by setting `subset_accuracy=True`.
    """

    tp: jax.Array
    fp: jax.Array
    tn: jax.Array
    fn: jax.Array

    mode: jm.DataType
    average: jm.AverageMethod
    mdmc_average: jm.MDMCAverageMethod
    num_classes: int | None
    threshold: float
    multiclass: bool | None
    ignore_index: int | None
    top_k: int | None
    subset_accuracy: bool

    @jt.jaxtyped(typechecker=beartype.beartype)
    def __init__(
        self,
        threshold: float = 0.5,
        num_classes: int | None = None,
        *,
        average: str | jm.AverageMethod = jm.AverageMethod.MICRO,
        mdmc_average: str | jm.MDMCAverageMethod = jm.MDMCAverageMethod.GLOBAL,
        ignore_index: int | None = None,
        top_k: int | None = None,
        multiclass: bool | None = None,
        subset_accuracy: bool = False,
        mode: str | jm.DataType = jm.DataType.MULTICLASS,
        tp: jax.Array = None,
        fp: jax.Array = None,
        tn: jax.Array = None,
        fn: jax.Array = None,
    ):
        """Init function.

        Args:
            subset_accuracy (bool, optional): Defaults to False.
            mode (Union[str, jm.DataType], optional): , defaults to
                jm.DataType.MULTICLASS.
            threshold (float, optional): Threshold for transforming
                probability or logit predictions to binary (0,1)
                predictions, defaults to 0.5.
            num_classes (Optional[int], optional): Number of classes.
                Necessary for `'macro'`, `'weighted'` and `None` average
                methods. in the case of binary or multi-label inputs.
                Default value of 0.5 corresponds to input being
                probabilities, defaults to None.
            average (Union[str, jm.AverageMethod], optional): Defines
                the reduction that is applied. Should be one of the
                following:

                - `'micro'` [default]: Calculate the metric globally, across all samples and
                  classes.
                - `'macro'`: Calculate the metric for each class separately, and average the
                metrics across classes (with equal weights for each class).
                - `'weighted'`: Calculate the metric for each class separately, and average the
                metrics across classes, weighting each class by its support (`tp + fn`).
                - `'none'` or `None`: Calculate the metric for each class separately, and return
                the metric for every class.
                - `'samples'`: Calculate the metric for each sample, and average the metrics
                across samples (with equal weights for each sample).

                .. note:: What is considered a sample in the multidimensional multi-class case
                depends on the value of `mdmc_average`.

                .. note:: If `'none'` and a given class doesn't occur in the `preds` or `target`,
                the value for the class will be `nan`, defaults to jm.AverageMethod.MICRO.
            mdmc_average (Union[str, jm.MDMCAverageMethod], optional):
                Defines how averaging is done for multidimensional
                multi-class inputs (on top of the `average` parameter).
                Should be one of the following:

                - `None` [default]: Should be left unchanged if your data is not multidimensional
                multi-class.

                - `'samplewise'`: In this case, the statistics are computed separately for each
                sample on the `N` axis, and then averaged over samples.
                The computation for each sample is done by treating the flattened extra axes `...`
                (see :ref:`references/modules:input types`) as the `N` dimension within the sample,
                and computing the metric for the sample based on that.

                - `'global'`: In this case the `N` and `...` dimensions of the inputs
                (see :ref:`references/modules:input types`)
                are flattened into a new `N_X` sample axis, i.e. the inputs are treated as if they
                were `(N_X, C)`. From here on the `average` parameter applies as usual, defaults to
                jm.MDMCAverageMethod.GLOBAL.
            ignore_index (Optional[int], optional): Integer specifying a
                target class to ignore. If given, this class index does
                not contribute to the returned score, regardless of
                reduction method. If an index is ignored, and
                `average=None` or `'none'`, the score for the ignored
                class will be returned as `nan`, defaults to None.
            top_k (Optional[int], optional): Number of highest
                probability or logit score predictions considered to
                find the correct label,  relevant only for
                (multidimensional) multi-class inputs. The default value
                (`None`) will be interpreted as 1 for these inputs.
                Should be left at default (`None`) for all other types
                of inputs, defaults to None.
            multiclass (Optional[bool], optional): Used only in certain
                special cases, where you want to treat inputs as a
                different type than what they appear to be. See the
                parameter's, defaults to None.
            fn (jax.Array, optional): Only used when merging, defaults
                to None.
            tn (jax.Array, optional): Only used when merging, defaults
                to None.
            fp (jax.Array, optional): Only used when merging, defaults
                to None.
            tp (jax.Array, optional): Only used when merging, defaults
                to None.

        Raises:
            ValueError: If `top_k` is not an `integer` larger than `0`.
            ValueError: If `average` is none of `"micro"`, `"macro"`,
                `"weighted"`, `"samples"`, `"none"`, `None`.
            ValueError: If two different input modes are provided, e.g.
                using `multi-label` with `multi-class`.
            ValueError: If `top_k` parameter is set for `multi-label`
                inputs.
        """
        if isinstance(average, str):
            average = jm.AverageMethod[average.upper()]

        if isinstance(mode, str):
            mode = jm.DataType[mode.upper()]

        if isinstance(mdmc_average, str):
            mdmc_average = jm.MDMCAverageMethod[mdmc_average.upper()]

        average = (
            jm.AverageMethod.MACRO
            if average in [jm.AverageMethod.WEIGHTED, jm.AverageMethod.NONE]
            else average
        )

        if average == jm.AverageMethod.MACRO and (not num_classes or num_classes < 1):
            raise ValueError(
                "When you set `reduce` as 'macro', you have to provide the number of classes."
            )

        if top_k is not None and top_k <= 0:
            raise ValueError(f"The `top_k` should be an integer larger than 0, got {top_k}")

        if (
            num_classes
            and ignore_index is not None
            and (not 0 <= ignore_index < num_classes or num_classes == 1)
        ):
            raise ValueError(
                f"The `ignore_index` {ignore_index} is not valid for inputs "
                "with {num_classes} classes"
            )

        if mdmc_average == "samplewise":
            raise ValueError(f"The `mdmc_average` method '{mdmc_average}' is not yet supported.")

        # Params
        self.average = average
        self.mdmc_average = mdmc_average
        self.num_classes = num_classes
        self.threshold = threshold
        self.multiclass = multiclass
        self.ignore_index = ignore_index
        self.top_k = top_k
        self.subset_accuracy = subset_accuracy
        self.mode = mode

        if tp is None:
            self.__dict__.update(self._initial_values())
        else:
            self.tp = tp
            self.fp = fp
            self.tn = tn
            self.fn = fn

    def _initial_values(self) -> dict[str, jax.Array]:
        """Initial values."""
        # nodes
        zeros_shape: list[int]
        if self.average == jm.AverageMethod.MICRO:
            zeros_shape = []
        elif self.average == jm.AverageMethod.MACRO:
            if self.num_classes is None:
                raise ValueError(
                    "The `num_classes` parameter must be defined to use `average='macro'`."
                )
            zeros_shape = [self.num_classes]
        else:
            raise ValueError(f'Wrong reduce="{self.average}"')

        initial_value = jnp.zeros(zeros_shape, dtype=jnp.uint32)

        return dict(
            tp=initial_value,
            fp=initial_value,
            tn=initial_value,
            fn=initial_value,
        )

    @override
    def empty(self) -> "Accuracy":
        """Empty function."""
        return Accuracy(
            threshold=self.threshold,
            num_classes=self.num_classes,
            average=self.average,
            mdmc_average=self.mdmc_average,
            ignore_index=self.ignore_index,
            top_k=self.top_k,
            multiclass=self.multiclass,
            subset_accuracy=self.subset_accuracy,
            mode=self.mode,
        )

    @override
    def create(
        # pylint: disable=arguments-differ
        self,
        preds: jax.Array,
        target: jax.Array,
    ) -> "Accuracy":
        """Updates Accuracy metric state.

        Args:
            preds (jax.Array): Predictions from model (logits,
                probabilities, or target).
            target (jax.Array): Ground truth target.

        Example:

            ``python
            accuracy = Accuracy().reset()

            accuracy = accuracy.update(preds=preds, target=target)
        """

        tp, fp, tn, fn = jm.stat_scores_update(
            preds,
            target,
            intended_mode=self.mode,
            average_method=self.average,
            mdmc_average_method=self.mdmc_average,
            threshold=self.threshold,
            num_classes=self.num_classes,
            top_k=self.top_k,
            multiclass=self.multiclass,
        )

        return dataclasses.replace(self, tp=tp, fp=fp, tn=tn, fn=fn)

    @override
    def merge(self, other: "Accuracy") -> "Accuracy":
        """Merge function."""
        return dataclasses.replace(
            self,
            tp=self.tp + other.tp,
            fp=self.fp + other.fp,
            tn=self.tn + other.tn,
            fn=self.fn + other.fn,
        )

    @override
    def compute(self) -> jax.Array:
        """Computes accuracy based on inputs passed in to `update` previously.

        :return s: The accuracy score.
        :rtype s: jax.Array
        """
        # if self.mode is None:
        #     raise RuntimeError("You have to have determined mode.")

        return jm.accuracy_compute(
            self.tp,
            self.fp,
            self.tn,
            self.fn,
            self.average,
            self.mdmc_average,
            self.mode,
        )



"""REAX: A simple training framework for JAX-based projects"""

from . import (
    _engine,
    data,
    exceptions,
    hooks,
    listeners,
    metrics,
    modules,
    optimizers,
    random,
    saving,
    stages,
    strategies,
    testing,
    training,
    types,
)
from ._engine import *
from .data import DataLoader, DataModule, ReaxDataLoader
from .hooks import *
from .loggers import Logger
from .metrics import Metric
from .modules import *
from .optimizers import *
from .random import *
from .random import seed_everything
from .saving import *
from .stages import Stage
from .strategies import *
from .training import *

__all__ = (
    _engine.__all__
    + modules.__all__
    + hooks.__all__
    + optimizers.__all__
    + random.__all__
    + saving.__all__
    + strategies.__all__
    + training.__all__
    + data.__all__
    # Modules
    + (
        "data",
        "exceptions",
        "stages",
        "listeners",
        "metrics",
        "strategies",
        "testing",
        "types",
    )
    # Classes/functions/variables
    + (
        "Metric",
        "DataLoader",
        "DataModule",
        "ReaxDataLoader",
        "Logger",
        "seed_everything",
        "Stage",
    )
)

__version__ = "0.6.2"



from . import _types, collate, fetchers, samplers

__all__ = ("GenericDataLoader",)


class GenericDataLoader(_types.DataLoader):
    def __init__(
        self,
        dataset: _types.Dataset,
        batch_size: int = 1,
        shuffle: bool = False,
        collate_fn: _types.CollateFn | None = None,
    ):
        """Init function."""
        self._batch_size = batch_size
        self._dataset = dataset
        self._sampler = samplers.create_sampler(
            dataset,
            batch_size=batch_size,
            shuffle=shuffle,
        )

        if collate_fn is None:
            collate_fn = collate.get_default_collator().collate

        self._fetcher = fetchers.create_fetcher(dataset, collate_fn=collate_fn)

    @property
    def batch_size(self) -> int:
        """Batch size."""
        return self._batch_size

    def __iter__(self):
        """Iter function."""
        for indices in self._sampler:
            yield self._fetcher.fetch(indices)



# Copyright (C) 2024  Martin Uhrin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Most of this file is covered by the following license.  To find what has been modified you
# can perform a diff with the file at:
# https://github.com/Lightning-AI/pytorch-lightning/blob/0324a20f00235c7a10a235a44326811ba42b6ae4/src/lightning/pytorch/callbacks/progress/rich_progress.py
#
# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Generator
import dataclasses
import datetime
import math
from typing import TYPE_CHECKING, Any, Final, cast

from lightning_utilities.core.imports import RequirementCache
from typing_extensions import override

from reax import stages

from . import progress_bar

if TYPE_CHECKING:
    import reax

__all__ = ("RichProgressBar",)

_RICH_AVAILABLE = RequirementCache("rich>=10.2.2")

if _RICH_AVAILABLE:  # noqa: C901
    import rich
    from rich import style as style_
    from rich import text as text_
    import rich.progress
    import rich.progress_bar

    class CustomBarColumn(rich.progress.BarColumn):
        """Overrides ``BarColumn`` to provide support for dataloaders that do not define a size
        (infinite size) such as ``IterableDataset``.
        """

        def render(self, task: rich.progress.Task) -> rich.progress_bar.ProgressBar:
            """Gets a progress bar widget for a task."""
            assert task.total is not None
            assert task.remaining is not None
            return rich.progress_bar.ProgressBar(
                total=max(0, task.total),
                completed=max(0, task.completed),
                width=None if self.bar_width is None else max(1, self.bar_width),
                pulse=not task.started or not math.isfinite(task.remaining),
                animation_time=task.get_time(),
                style=self.style,
                complete_style=self.complete_style,
                finished_style=self.finished_style,
                pulse_style=self.pulse_style,
            )

    @dataclasses.dataclass
    class CustomInfiniteTask(rich.progress.Task):
        """Overrides ``Task`` to define an infinite task.

        This is useful for datasets that do not define a size (infinite size) such as
        ``IterableDataset``.
        """

        @property
        def time_remaining(self) -> float | None:
            return None

    class CustomProgress(rich.progress.Progress):
        """Overrides ``Progress`` to support adding tasks that have an infinite total size."""

        def add_task(
            self,
            description: str,
            start: bool = True,
            total: float | None = 100.0,
            completed: int = 0,
            visible: bool = True,
            **fields: Any,
        ) -> rich.progress.TaskID:
            assert total is not None
            if not math.isfinite(total):
                task = CustomInfiniteTask(
                    self._task_index,
                    description,
                    total,
                    completed,
                    visible=visible,
                    fields=fields,
                    _get_time=self.get_time,
                    _lock=self._lock,
                )
                return self.add_custom_task(task)
            return super().add_task(description, start, total, completed, visible, **fields)

        def add_custom_task(
            self, task: CustomInfiniteTask, start: bool = True
        ) -> rich.progress.TaskID:
            with self._lock:
                self._tasks[self._task_index] = task
                if start:
                    self.start_task(self._task_index)
                new_task_index = self._task_index
                self._task_index = rich.progress.TaskID(int(self._task_index) + 1)
            self.refresh()
            return new_task_index

    class CustomTimeColumn(rich.progress.ProgressColumn):
        # Only refresh twice a second to prevent jitter
        max_refresh = 0.5

        def __init__(self, style: str | style_.Style) -> None:
            self.style = style
            super().__init__()

        def render(self, task: rich.progress.Task) -> text_.Text:
            elapsed = task.finished_time if task.finished else task.elapsed
            remaining = task.time_remaining
            elapsed_delta = (
                "-:--:--" if elapsed is None else str(datetime.timedelta(seconds=int(elapsed)))
            )
            remaining_delta = (
                "-:--:--" if remaining is None else str(datetime.timedelta(seconds=int(remaining)))
            )
            return text_.Text(f"{elapsed_delta} • {remaining_delta}", style=self.style)

    class BatchesProcessedColumn(rich.progress.ProgressColumn):
        def __init__(self, style: str | style_.Style):
            self.style = style
            super().__init__()

        def render(self, task: rich.progress.Task) -> rich.console.RenderableType:
            total = task.total if task.total != float("inf") else "--"
            return text_.Text(f"{int(task.completed)}/{total}", style=self.style)

    class ProcessingSpeedColumn(rich.progress.ProgressColumn):
        def __init__(self, style: str | style_.Style):
            self.style = style
            super().__init__()

        def render(self, task: rich.progress.Task) -> rich.console.RenderableType:
            task_speed = f"{task.speed:>.2f}" if task.speed is not None else "0.00"
            return text_.Text(f"{task_speed}it/s", style=self.style)

    class MetricsTextColumn(rich.progress.ProgressColumn):
        """A column containing text."""

        def __init__(
            self,
            trainer: "reax.Trainer",
            style: str | style_.Style,
            text_delimiter: str,
            metrics_format: str,
        ):
            self._trainer = trainer
            self._tasks: dict[int | rich.progress.TaskID, Any] = {}
            self._current_task_id = 0
            self._metrics: dict[str | style_.Style, Any] = {}
            self._style = style
            self._text_delimiter = text_delimiter
            self._metrics_format = metrics_format
            super().__init__()

        def update(self, metrics: dict[Any, Any]) -> None:
            # Called when metrics are ready to be rendered.
            # This is to prevent render from causing deadlock issues by requesting metrics
            # in separate threads.
            self._metrics = metrics

        def render(self, task: rich.progress.Task) -> text_.Text:
            assert isinstance(self._trainer.progress_bar_listener, RichProgressBar)
            if (
                not isinstance(self._trainer.stage, stages.Fit)
                # TODO: or self._trainer.sanity_checking
                or self._trainer.progress_bar_listener.train_progress_bar_id != task.id
            ):
                return text_.Text()

            if isinstance(self._trainer.stage, stages.Fit) and task.id not in self._tasks:
                self._tasks[task.id] = "None"
                if self._renderable_cache:
                    self._current_task_id = cast(rich.progress.TaskID, self._current_task_id)
                    self._tasks[self._current_task_id] = self._renderable_cache[
                        self._current_task_id
                    ][1]
                self._current_task_id = task.id

            if isinstance(self._trainer.stage, stages.Fit) and task.id != self._current_task_id:
                return self._tasks[task.id]

            metrics_texts = self._generate_metrics_texts()
            text = self._text_delimiter.join(metrics_texts)
            return text_.Text(text, justify="left", style=self._style)

        def _generate_metrics_texts(self) -> Generator[str, None, None]:
            for name, value in self._metrics.items():
                if not isinstance(value, str):
                    value = f"{value:{self._metrics_format}}"
                yield f"{name}: {value}"


@dataclasses.dataclass
class RichProgressBarTheme:
    """Styles to associate to different base components.

    Args:
        description: Style for the progress bar description. For eg., Epoch x, Testing, etc.
        progress_bar: Style for the bar in progress.
        progress_bar_finished: Style for the finished progress bar.
        progress_bar_pulse: Style for the progress bar when `IterableDataset` is being processed.
        batch_progress: Style for the progress tracker (i.e 10/50 batches completed).
        time: Style for the processed time and estimate time remaining.
        processing_speed: Style for the speed of the batches being processed.
        metrics: Style for the metrics

    https://rich.readthedocs.io/en/stable/style.html

    """

    description: str | style_.Style = ""
    progress_bar: str | style_.Style = "#6206E0"
    progress_bar_finished: str | style_.Style = "#6206E0"
    progress_bar_pulse: str | style_.Style = "#6206E0"
    batch_progress: str | style_.Style = ""
    time: str | style_.Style = "dim"
    processing_speed: str | style_.Style = "dim underline"
    metrics: str | style_.Style = "italic"
    metrics_text_delimiter: str = " "
    metrics_format: str = ".3f"


class RichProgressBar(progress_bar.ProgressBar):
    """Create a progress bar with `rich text formatting <https://github.com/Textualize/rich>`_.

    Install it with pip:

    .. code-block:: bash

        pip install rich

    .. code-block:: python

        import reax

        trainer = reax.Trainer(callbacks=reax.listeners.progress.RichProgressBar())

    Args:
        refresh_rate: Determines at which rate (in number of batches) the progress bars get updated.
            Set it to ``0`` to disable the display.
        leave: Leaves the finished progress bar in the terminal at the end of the epoch.
            Default: False
        theme: Contains styles used to stylize the progress bar.
        console_kwargs: Args for constructing a `Console`

    Raises:
        ModuleNotFoundError:
            If required `rich` package is not installed on the device.

    Note:
        PyCharm users will need to enable “emulate terminal” in output console option in
        run/debug configuration to see styled output.
        Reference: https://rich.readthedocs.io/en/latest/introduction.html#requirements

    """

    # pylint: disable=too-many-public-methods

    def __init__(
        self,
        refresh_rate: int = 1,
        leave: bool = False,
        theme: RichProgressBarTheme = RichProgressBarTheme(),
        console_kwargs: dict[str, Any] | None = None,
    ) -> None:
        if not _RICH_AVAILABLE:
            raise ModuleNotFoundError(
                "`RichProgressBar` requires `rich` >= 10.2.2. Install it by running "
                "`pip install -U rich`."
            )

        super().__init__()
        # Params
        self._refresh_rate: Final[int] = refresh_rate
        self._leave: Final[bool] = leave

        # State
        self._console: rich.Console | None = None
        self._console_kwargs = console_kwargs or {}
        self._enabled: bool = True
        self.progress: CustomProgress | None = None
        self.train_progress_bar_id: rich.progress.TaskID | None
        self.val_sanity_progress_bar_id: rich.progress.TaskID | None = None
        self.val_progress_bar_id: rich.progress.TaskID | None
        self.test_progress_bar_id: rich.progress.TaskID | None
        self.predict_progress_bar_id: rich.progress.TaskID | None
        self._reset_progress_bar_ids()
        self._metric_component: MetricsTextColumn | None = None
        self._progress_stopped: bool = False
        self.theme = theme

    @property
    def refresh_rate(self) -> float:
        return self._refresh_rate

    @property
    def is_enabled(self) -> bool:
        return self._enabled and self.refresh_rate > 0

    @property
    def is_disabled(self) -> bool:
        return not self.is_enabled

    @property
    def train_progress_bar(self) -> rich.progress.Task:
        assert self.progress is not None
        assert self.train_progress_bar_id is not None
        return self.progress.tasks[self.train_progress_bar_id]

    @property
    def val_sanity_check_bar(self) -> rich.progress.Task:
        assert self.progress is not None
        assert self.val_sanity_progress_bar_id is not None
        return self.progress.tasks[  # pylint: disable=invalid-sequence-index
            self.val_sanity_progress_bar_id
        ]

    @property
    def val_progress_bar(self) -> rich.progress.Task:
        assert self.progress is not None
        assert self.val_progress_bar_id is not None
        return self.progress.tasks[self.val_progress_bar_id]

    @property
    def test_progress_bar(self) -> rich.progress.Task:
        assert self.progress is not None
        assert self.test_progress_bar_id is not None
        return self.progress.tasks[self.test_progress_bar_id]

    @override
    def disable(self) -> None:
        self._enabled = False

    @override
    def enable(self) -> None:
        self._enabled = True

    def _init_progress(self, trainer: "reax.Trainer") -> None:
        if self.is_enabled and (self.progress is None or self._progress_stopped):
            self._reset_progress_bar_ids()
            rich.reconfigure(**self._console_kwargs)
            self._console = rich.get_console()
            if self._console.is_terminal:
                self._console.clear_live()
            self._metric_component = MetricsTextColumn(
                trainer,
                self.theme.metrics,
                self.theme.metrics_text_delimiter,
                self.theme.metrics_format,
            )
            self.progress = CustomProgress(
                *self._configure_columns(trainer),
                self._metric_component,
                auto_refresh=False,
                disable=self.is_disabled,
                console=self._console,
            )
            self.progress.start()
            # progress has started
            self._progress_stopped = False

    def refresh(self) -> None:
        if self.progress:
            self.progress.refresh()

    @override
    def on_train_start(self, trainer: "reax.Trainer", /, *_) -> None:
        self._init_progress(trainer)

    @override
    def on_predict_start(self, trainer: "reax.Trainer", /, *_) -> None:
        self._init_progress(trainer)

    @override
    def on_test_start(self, trainer: "reax.Trainer", /, *_) -> None:
        self._init_progress(trainer)

    @override
    def on_validation_start(self, trainer: "reax.Trainer", /, *_) -> None:
        self._init_progress(trainer)

    @override
    def on_sanity_check_start(self, trainer: "reax.Trainer", *_) -> None:
        self._init_progress(trainer)

    @override
    def on_train_epoch_start(self, trainer: "reax.Trainer", stage: "reax.stages.Train", /) -> None:
        if self.is_disabled:
            return

        total_batches = stage.max_batches
        train_description = self._get_train_description(trainer.current_epoch, stage)

        if self.train_progress_bar_id is not None and self._leave:
            self._stop_progress()
            self._init_progress(trainer)
        if self.progress is not None:
            if self.train_progress_bar_id is None:
                self.train_progress_bar_id = self._add_task(total_batches, train_description)
            else:
                self.progress.reset(
                    self.train_progress_bar_id,
                    total=total_batches,
                    description=train_description,
                    visible=True,
                )

        self.refresh()

    @override
    def on_train_batch_end(
        self,
        trainer: "reax.Trainer",
        stage: "reax.stages.Train",
        _outputs: Any,
        _batch: Any,
        batch_idx: int,
        /,
    ) -> None:
        self._update(self.train_progress_bar_id, batch_idx + 1)
        self._update_metrics(trainer, stage)
        self.refresh()

    @override
    def on_train_epoch_end(self, trainer: "reax.Trainer", stage: "reax.stages.Train", /) -> None:
        self._update_metrics(trainer, stage)

    @override
    def on_sanity_check_end(self, *_) -> None:
        if self.progress is not None:
            assert self.val_sanity_progress_bar_id is not None
            self.progress.update(self.val_sanity_progress_bar_id, advance=0, visible=False)
        self.refresh()

    @override
    def on_validation_epoch_start(
        self, _trainer: "reax.Trainer", stage: "reax.stages.Validate", /
    ) -> None:
        if self.is_disabled:  # TODO or not self.has_dataloader_changed(dataloader_idx):
            return

        assert self.progress is not None

        # TODO:
        # if trainer.sanity_checking:
        #     if self.val_sanity_progress_bar_id is not None:
        #         self.progress.update(self.val_sanity_progress_bar_id, advance=0, visible=False)
        #
        #     self.val_sanity_progress_bar_id = self._add_task(
        #         stage.max_batches, self.sanity_check_description, visible=False
        #     )
        # else:
        if self.val_progress_bar_id is not None:
            self.progress.update(self.val_progress_bar_id, advance=0, visible=False)

        # TODO: remove old tasks when new ones are created
        self.val_progress_bar_id = self._add_task(
            stage.max_batches, self.validation_description, visible=False
        )

        self.refresh()

    def _add_task(
        self, total_batches: int | float, description: str, visible: bool = True
    ) -> rich.progress.TaskID:
        assert self.progress is not None
        return self.progress.add_task(
            f"[{self.theme.description}]{description}" if self.theme.description else description,
            total=total_batches,
            visible=visible,
        )

    def _update(
        self, progress_bar_id: rich.progress.TaskID | None, current: int, visible: bool = True
    ) -> None:
        if self.progress is not None and self.is_enabled:
            assert progress_bar_id is not None
            total = self.progress.tasks[progress_bar_id].total
            assert total is not None
            if not self._should_update(current, total):
                return

            leftover = current % self.refresh_rate
            advance = leftover if (current == total and leftover != 0) else self.refresh_rate
            self.progress.update(progress_bar_id, advance=advance, visible=visible)
            self.refresh()

    def _should_update(self, current: int, total: int | float) -> bool:
        return current % self.refresh_rate == 0 or current == total

    @override
    def on_validation_epoch_end(
        self, trainer: "reax.Trainer", _stage: "reax.stages.Validate", /
    ) -> None:
        if (
            self.is_enabled
            and self.val_progress_bar_id is not None
            and isinstance(trainer.stage, stages.Fit)
        ):
            assert self.progress is not None
            self.progress.update(self.val_progress_bar_id, advance=0, visible=False)
            self.refresh()

    # @override
    # def on_validation_end(self, trainer: "reax.Trainer",
    #       stage: "reax.stages.Validate", /) -> None:
    #     if isinstance(trainer.stage, stages.Fit):
    #         self._update_metrics(trainer, stage)
    #     self.reset_dataloader_idx_tracker()
    #
    # @override
    # def on_test_end(self, trainer: "reax.Trainer", stage: "reax.stages.Test", /) -> None:
    #     self.reset_dataloader_idx_tracker()
    #
    # @override
    # def on_predict_end(self, trainer: "reax.Trainer", stage: "reax.stages.Predict", /) -> None:
    #     self.reset_dataloader_idx_tracker()

    @override
    def on_test_epoch_start(self, _trainer: "reax.Trainer", stage: "reax.stages.Test", /) -> None:
        if self.is_disabled:  # or not self.has_dataloader_changed(dataloader_idx):
            return

        if self.test_progress_bar_id is not None:
            assert self.progress is not None
            self.progress.update(self.test_progress_bar_id, advance=0, visible=False)
        self.test_progress_bar_id = self._add_task(stage.max_batches, self.test_description)
        self.refresh()

    @override
    def on_predict_epoch_start(
        self, _trainer: "reax.Trainer", stage: "reax.stages.Predict", /
    ) -> None:
        if self.is_disabled:  # or not self.has_dataloader_changed(dataloader_idx):
            return

        if self.predict_progress_bar_id is not None:
            assert self.progress is not None
            self.progress.update(self.predict_progress_bar_id, advance=0, visible=False)
        self.predict_progress_bar_id = self._add_task(stage.max_batches, self.predict_description)
        self.refresh()

    @override
    def on_validation_batch_end(
        self,
        _trainer: "reax.Trainer",
        _stage: "reax.stages.Validate",
        _outputs: Any,
        _batch: Any,
        batch_idx: int,
        /,
    ) -> None:
        if self.is_disabled:
            return

        # TODO: if trainer.sanity_checking:
        #     self._update(self.val_sanity_progress_bar_id, batch_idx + 1)
        if self.val_progress_bar_id is not None:
            self._update(self.val_progress_bar_id, batch_idx + 1)

        self.refresh()

    @override
    def on_test_batch_end(
        self,
        _trainer: "reax.Trainer",
        _stage: "reax.stages.Test",
        _outputs: Any,
        _batch: Any,
        batch_idx: int,
        /,
        # dataloader_idx: int = 0,
    ) -> None:
        if self.is_disabled:
            return
        assert self.test_progress_bar_id is not None
        self._update(self.test_progress_bar_id, batch_idx + 1)
        self.refresh()

    @override
    def on_predict_batch_end(
        self,
        _trainer: "reax.Trainer",
        _stage: "reax.stages.Predict",
        _outputs: Any,
        _batch: Any,
        batch_idx: int,
        /,
    ) -> None:
        if self.is_disabled:
            return
        assert self.predict_progress_bar_id is not None
        self._update(self.predict_progress_bar_id, batch_idx + 1)
        self.refresh()

    def _get_train_description(self, current_epoch: int, stage: "reax.stages.Train") -> str:
        train_description = f"Epoch {current_epoch}"
        if stage.parent is not None and stage.parent.max_iters is not None:
            max_epochs = stage.parent.max_iters
            train_description += f"/{max_epochs - 1}"
        if len(self.validation_description) > len(train_description):
            # Padding is required to avoid flickering due of uneven lengths of "Epoch X"
            # and "Validation" Bar description
            train_description = f"{train_description:{len(self.validation_description)}}"
        return train_description

    def _stop_progress(self) -> None:
        if self.progress is not None:
            self.progress.stop()
            # # signals for progress to be re-initialized for next stages
            self._progress_stopped = True

    def _reset_progress_bar_ids(self) -> None:
        self.train_progress_bar_id = None
        self.val_sanity_progress_bar_id = None
        self.val_progress_bar_id = None
        self.test_progress_bar_id = None
        self.predict_progress_bar_id = None

    def _update_metrics(self, trainer: "reax.Trainer", stage: "reax.stages.Stage") -> None:
        metrics = self.get_metrics(trainer, stage)
        if self._metric_component:
            self._metric_component.update(metrics)

    @override
    def teardown(self, /, *_) -> None:
        self._stop_progress()

    @override
    def on_exception(self, /, *_) -> None:
        self._stop_progress()

    def _configure_columns(self, _trainer: "reax.Trainer", /) -> list:
        if not _RICH_AVAILABLE:
            return []

        # pylint: disable=possibly-used-before-assignment
        return [
            rich.progress.TextColumn("[progress.description]{task.description}"),
            CustomBarColumn(
                complete_style=self.theme.progress_bar,
                finished_style=self.theme.progress_bar_finished,
                pulse_style=self.theme.progress_bar_pulse,
            ),
            BatchesProcessedColumn(style=self.theme.batch_progress),
            CustomTimeColumn(style=self.theme.time),
            ProcessingSpeedColumn(style=self.theme.processing_speed),
        ]

    def __getstate__(self) -> dict:
        state = self.__dict__.copy()
        # both the console and progress object can hold thread lock objects that are not pickleable
        state["progress"] = None
        state["_console"] = None
        return state



import argparse
from collections.abc import Callable, Mapping
import os
from typing import Any, Final, Optional, Union

import fsspec
import jax.typing
from lightning_utilities.core import rank_zero
import numpy as np
import pandas as pd
from typing_extensions import override

from reax import saving

from . import _utils, logger

__all__ = ("PandasLogger",)

DEFAULT_FORMAT = "json"
Path = Union[str, bytes, os.PathLike]


class PandasLogger(logger.WithDdp["ExperimentWriter"], logger.Logger):
    """Log results in a pandas dataframe."""

    LOGGER_JOIN_CHAR: Final[str] = "-"

    def __init__(
        self,
        save_dir: Path | None = None,
        name: str | None = "reax_logs",
        *,
        fmt: str = DEFAULT_FORMAT,
        version: int | str | None = None,
        prefix: str = "",
        flush_logs_every_n_steps: int = 100,
    ):
        super().__init__()
        self._root_dir = None if save_dir is None else os.fspath(save_dir)
        self._name = name or ""
        self._fmt = fmt
        self._version = version
        self._prefix = prefix
        self._flush_logs_every_n_step = flush_logs_every_n_steps

        self._exp: Optional["ExperimentWriter"] = None
        self._fs: fsspec.AbstractFileSystem = fsspec.url_to_fs(self.root_dir)[0]

    @property
    @override
    def name(self) -> str:
        """Name function."""
        return self._name

    @property
    @override
    def version(self) -> int | str:
        """Version function."""
        if self._version is None:
            self._version = self._get_next_version()
        return self._version

    @property
    @override
    def root_dir(self) -> str | None:
        """Root dir."""
        return self._root_dir

    @property
    @override
    def log_dir(self) -> str | None:
        """Log dir."""
        if self.root_dir is None:
            return None

        # Use this scheme to generate some kind of standard logging path
        version = self.version if isinstance(self.version, str) else f"version_{self.version}"
        return os.path.join(self._root_dir, self.name, version)

    @property
    def dataframe(self) -> pd.DataFrame:
        """Get the dataframe from the current experiment."""
        return self.experiment.dataframe

    @property
    def _experiment(self) -> "ExperimentWriter":
        """Experiment function."""
        if self._exp is None:
            self._exp = ExperimentWriter(log_dir=self.log_dir, fmt=self._fmt)

        return self._exp

    @property
    @override
    def save_dir(self) -> str | None:
        """Save dir."""
        return self._root_dir

    @override
    def _log_metrics(self, metrics: Mapping[str, float], step: int | None = None) -> None:
        """Log metrics."""
        metrics = _utils.add_prefix(metrics, self._prefix, self.LOGGER_JOIN_CHAR)

        if step is None:
            step = len(self.experiment.metrics)

        self.experiment.log_metrics(metrics, step)
        if (step + 1) % self._flush_logs_every_n_step == 0:
            self.save()

    @override
    def _log_hyperparams(
        # pylint: disable=arguments-differ
        self,
        params: dict[str, Any] | argparse.Namespace,
    ) -> None:
        """Log hyperparams."""
        params = _utils.convert_params(params)
        self.experiment.log_hparams(params)

    @override
    def _save(self) -> None:
        """Save function."""
        self.experiment.save()

    @override
    def _finalize(self, _status: str) -> None:
        """Finalize function."""
        if self._experiment is not None:
            return

        self.save()

    def _get_next_version(self) -> int:
        """Get next version."""
        if self.root_dir is None:
            return 0

        version_root = os.path.join(self._root_dir, self.name)

        if not self._fs.isdir(version_root):
            # Doesn't exist yet
            return 0
        try:
            listdir_info = self._fs.listdir(version_root)
        except OSError:
            return 0

        existing_versions = []
        for info in listdir_info:
            name = info["name"]
            name = os.path.join(version_root, name)
            basename = os.path.basename(name)
            if self._fs.isdir(name) and basename.startswith("version_"):
                dir_ver = basename.split("_")[1].replace("/", "")
                if dir_ver.isdigit():
                    existing_versions.append(int(dir_ver))

        if len(existing_versions) == 0:
            return 0

        return max(existing_versions) + 1

    @staticmethod
    def _sanitize_params(params: dict[str, Any]) -> dict[str, Any]:
        """Sanitize params."""
        params = _utils.sanitize_params(params)
        # logging of arrays with dimension > 1 is not supported, sanitize as string
        return {
            key: str(val) if hasattr(val, "ndim") and val.ndim > 1 else val
            for key, val in params.items()
        }


class ExperimentWriter:
    r"""Pandas experiment write."""

    DEFAULT_BASENAME = "metrics"
    HPARAMS_FILENAME = "hparams.yaml"

    def __init__(self, log_dir: str | None, fmt: str = DEFAULT_FORMAT):
        try:
            getattr(pd.DataFrame, f"to_{fmt}")
        except AttributeError:
            supported_formats = [
                name[len("to_") :] for name in dir(pd.DataFrame) if name.startswith("to_")
            ]
            raise ValueError(
                f"Format '{fmt}' is not supported.  Try one of {','.join(supported_formats)}'"
            ) from None

        self._log_dir = log_dir
        self._fmt = fmt

        self._rows: list[dict[str, float]] = []

        if self._log_dir is None:
            self._metrics_file_path = None
            self._fs = None
        else:
            filename = f"{self.DEFAULT_BASENAME}.{DEFAULT_FORMAT}"
            self._metrics_file_path = os.path.join(self._log_dir, filename)
            self._fs = fsspec.url_to_fs(log_dir)[0]
            self._check_log_dir_exists()
            self._fs.makedirs(self._log_dir, exist_ok=True)

        self.hparams: dict[str, Any] = {}

    @property
    def log_dir(self) -> str | None:
        """Log dir."""
        return self._log_dir

    @property
    def metrics(self):
        """Metrics function."""
        return self._rows

    @property
    def metrics_file_path(self) -> str | None:
        """Metrics file path."""
        return self._metrics_file_path

    @property
    def dataframe(self) -> pd.DataFrame:
        """Get the experiment's dataframe."""
        return pd.DataFrame(self._rows)

    def log_metrics(self, metrics: dict[str, float], step: int | None = None) -> None:
        """Log metrics."""

        def _handle_value(value: jax.Array | np.ndarray | Any) -> Any:
            """Handle value."""
            if isinstance(value, (jax.Array, np.ndarray)):
                return value.item()
            return value

        if step is None:
            step = len(self._rows)

        metrics = jax.tree.map(_handle_value, metrics)
        metrics["step"] = step
        self._rows.append(metrics)

    def log_hparams(self, params: dict[str, Any]) -> None:
        """Record hparams."""
        self.hparams.update(params)

    def save(self) -> None:
        """Save recorded metrics into files."""
        if self._log_dir is None:
            return

        hparams_file = os.path.join(self._log_dir, self.HPARAMS_FILENAME)
        saving.save_hparams_to_yaml(hparams_file, self.hparams, use_omegaconf=False)

        if not self._rows:
            return

        dframe = self.dataframe
        save_method: Callable = getattr(dframe, f"to_{self._fmt}")
        save_method(self._metrics_file_path)

    def _check_log_dir_exists(self) -> None:
        """Check log dir exists."""
        if self._fs.exists(self.log_dir) and self._fs.listdir(self.log_dir):
            rank_zero.rank_zero_warn(
                f"Experiment logs directory {self.log_dir} exists and is not empty."
                " Previous log files in this directory will be deleted when the new ones are saved!"
            )
            if self._fs.isfile(self.metrics_file_path):
                self._fs.rm_file(self.metrics_file_path)



# Copyright (C) 2024  Martin Uhrin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Most of this file is covered by the following license.  To find what has been modified you
# can perform a diff with the file at:
# https://github.com/Lightning-AI/pytorch-lightning/blob/0324a20f00235c7a10a235a44326811ba42b6ae4/src/lightning/pytorch/callbacks/early_stopping.py
#
# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Early Stopping
^^^^^^^^^^^^^^

Monitor a metric and stop training when it stops improving.

"""
from collections.abc import Callable
import logging
from typing import TYPE_CHECKING, Final, Literal

import beartype
import jax.numpy as jnp
import jax.typing
import jaxtyping as jt
from typing_extensions import override

from reax import hooks, stages
from reax.lightning import rank_zero

if TYPE_CHECKING:
    import reax

_LOGGER = logging.getLogger(__name__)

__all__ = ("EarlyStopping",)

MonitorOp = Callable[[jax.typing.ArrayLike, jax.typing.ArrayLike], jax.typing.ArrayLike]


class EarlyStopping(hooks.TrainerListener):
    r"""Monitor a metric and stop training when it stops improving.

    Args:
        monitor (str): Quantity to be monitored.
        min_delta (float, optional): Minimum change in the monitored
            quantity to qualify as an improvement, i.e. an absolute
            change of less than or equal to ``min_delta``, will count as
            no improvement, defaults to 0.0.
        patience (int, optional): Number of checks with no improvement
            after which training will be stopped. Under the default
            configuration, one check happens after every training epoch.
            However, the frequency of validation can be modified by
            setting various parameters on the ``Trainer``, for example
            ``check_val_every_n_epoch`` and ``val_check_interval``.

            .. note::

            It must be noted that the patience parameter counts the number of validation checks with
            no improvement, and not the number of training epochs. Therefore, with parameters
            ``check_val_every_n_epoch=10`` and ``patience=3``, the trainer will perform at least 40
            training epochs before being stopped, defaults to 3.
        verbose (bool, optional): Verbosity mode, defaults to False.
        mode (Literal["min", "max"], optional): One of ``'min'``,
            ``'max'``. In ``'min'`` mode, training will stop when the
            quantity monitored has stopped decreasing and in ``'max'``
            mode it will stop when the quantity monitored has stopped
            increasing, defaults to "min".
        strict (bool, optional): Whether to crash the training if
            `monitor` is not found in the validation metrics, defaults
            to True.
        check_finite (bool, optional): When set ``True``, stops training
            when the monitor becomes NaN or infinite, defaults to True.
        stopping_threshold (Optional[float], optional): Stop training
            immediately once the monitored quantity reaches this
            threshold, defaults to None.
        divergence_threshold (Optional[float], optional): Stop training
            as soon as the monitored quantity becomes worse than this
            threshold, defaults to None.
        check_on_train_epoch_end (bool, optional): Whether to run early
            stopping at the end of the training epoch. If this is
            ``False``, then the check runs at the end of the validation,
            defaults to None.
        log_rank_zero_only (bool, optional): When set ``True``, logs the
            status of the early stopping listener only for rank 0
            process, defaults to False.

    Raises:
        MisconfigurationException: If ``mode`` is none of ``"min"`` or
            ``"max"``.
        RuntimeError: If the metric ``monitor`` is not available.

    Example::

        >>> from reax import Trainer
        >>> from reax.listeners import EarlyStopping
        >>> early_stopping = EarlyStopping('val_loss')
        >>> trainer = Trainer(listeners=[early_stopping])
    """

    mode_dict: dict[str, MonitorOp] = {"min": jnp.less, "max": jnp.greater}
    order_dict = {"min": "<", "max": ">"}

    @jt.jaxtyped(typechecker=beartype.beartype)
    def __init__(
        self,
        monitor: str,
        *,
        min_delta: float = 0.0,
        patience: int = 3,
        verbose: bool = False,
        mode: Literal["min", "max"] = "min",
        strict: bool = True,
        check_finite: bool = True,
        stopping_threshold: float | None = None,
        divergence_threshold: float | None = None,
        check_on_train_epoch_end: bool = None,
        log_rank_zero_only: bool = False,
    ):
        # Params
        self._monitor: Final[str] = monitor
        self._patience: Final[int] = patience
        self._verbose: Final[bool] = verbose
        self._mode: Final[Literal["min", "max"]] = mode
        self._strict: Final[bool] = strict
        self._check_finite = check_finite
        self._stopping_threshold = stopping_threshold
        self._divergence_threshold = divergence_threshold
        self._wait_count = 0
        self.__check_on_train_epoch_end = check_on_train_epoch_end
        self._log_rank_zero_only = log_rank_zero_only
        self._min_delta: Final[float] = min_delta if mode == "max" else -min_delta

        # State
        self._check_on_train_epoch_end = check_on_train_epoch_end
        self._best_score = float("inf")
        self._wait_count: int = 0
        self._stopped_epoch = 0

    @property
    def monitor_op(self) -> MonitorOp:
        """Monitor op."""
        return self.mode_dict[self._mode]

    @property
    def best_score(self) -> float:
        """Best score."""
        return self._best_score

    @property
    def stopped_epoch(self) -> int:
        return self._stopped_epoch

    def _should_skip_check(self, trainer: "reax.Trainer") -> bool:
        """Should skip check."""
        return not isinstance(trainer.stage, stages.Fit)

    @override
    def on_fit_start(self, _trainer: "reax.Trainer", stage: "reax.stages.Fit", /) -> None:
        """On fit start."""
        if self.__check_on_train_epoch_end is None:
            # if the user runs validation multiple times per training epoch or multiple training
            # epochs without validation, then we run after validation instead of on train epoch end
            self._check_on_train_epoch_end = (
                stage.val_check_interval == 1.0 and stage.check_val_every_n_epoch == 1
            )

    @override
    def on_train_epoch_end(self, trainer: "reax.Trainer", stage: "reax.stages.Train", /) -> None:
        """On train epoch end."""
        if not self._check_on_train_epoch_end or self._should_skip_check(trainer):
            return
        self._run_early_stopping_check(trainer, stage)

    @override
    def on_validation_end(self, trainer: "reax.Trainer", stage: "reax.stages.Validate", /) -> None:
        """On validation end."""
        if self._check_on_train_epoch_end or self._should_skip_check(trainer):
            return
        self._run_early_stopping_check(trainer, stage)

    def _run_early_stopping_check(
        self, trainer: "reax.Trainer", stage: "reax.stages.EpochStage"
    ) -> None:
        """Check the early stopping condition and tell the Trainer to stop if needed."""
        logs = trainer.listener_metrics

        if stage.fast_dev_run or not self._validate_condition_metric(logs):
            # disable early_stopping with fast_dev_run or short circuit if metric not present
            return

        current = jnp.squeeze(logs[self._monitor])
        should_stop, reason = self._evaluate_stopping_criteria(current)

        # stop every ddp process if any process decides to stop
        # should_stop = trainer.strategy.reduce_boolean_decision(should_stop, all=False)
        trainer.should_stop = trainer.should_stop or should_stop
        if should_stop:
            self._stopped_epoch = trainer.current_epoch
        if reason and self._verbose:
            self._log_info(trainer, reason, self._log_rank_zero_only)

    def _validate_condition_metric(self, logs: dict[str, jax.typing.ArrayLike]) -> bool:
        """Validate condition metric."""
        monitor_val = logs.get(self._monitor)

        error_msg = (
            f"Early stopping conditioned on metric `{self._monitor}` which is not available."
            " Pass in or modify your `EarlyStopping` listener to use any of the following:"
            f' `{"`, `".join(list(logs.keys()))}`'
        )

        if monitor_val is None:
            if self._strict:
                raise RuntimeError(error_msg)
            if self._verbose > 0:
                rank_zero.rank_zero_warn(error_msg, category=RuntimeWarning)

            return False

        return True

    def _evaluate_stopping_criteria(self, current: jax.typing.ArrayLike) -> tuple[bool, str | None]:
        """Evaluate stopping criteria."""
        should_stop = False
        reason = None
        if self._check_finite and not jnp.isfinite(current):
            should_stop = True
            reason = (
                f"Monitored metric {self._monitor} = {current} is not finite."
                f" Previous best value was {self.best_score:.3f}. Signaling Trainer to stop."
            )
        elif self._stopping_threshold is not None and self.monitor_op(
            current, self._stopping_threshold
        ):
            should_stop = True
            reason = (
                f"Stopping threshold reached: {self._monitor} = {current} "
                f"{self.order_dict[self._mode]} {self._stopping_threshold}. "
                f"Signaling Trainer to stop."
            )
        elif self._divergence_threshold is not None and self.monitor_op(
            -current,
            -self._divergence_threshold,  # pylint: disable=invalid-unary-operand-type
        ):
            should_stop = True
            reason = (
                f"Divergence threshold reached: {self._monitor} = {current} "
                f"{self.order_dict[self._mode]} {self._divergence_threshold}."
                " Signaling Trainer to stop."
            )
        elif self.monitor_op(current - self._min_delta, self.best_score):
            should_stop = False
            reason = self._improvement_message(current)
            self._best_score = current
            self._wait_count = 0
        else:
            self._wait_count += 1
            if self._wait_count >= self._patience:
                should_stop = True
                reason = (
                    f"Monitored metric {self._monitor} did not improve in the last "
                    f"{self._wait_count} records. "
                    f"Best score: {self.best_score:.3f}. Signaling Trainer to stop."
                )

        return should_stop, reason

    def _improvement_message(self, current: jax.typing.ArrayLike) -> str:
        """Formats a log message that informs the user about an improvement in the monitored
        score.
        """
        if jnp.isfinite(self.best_score):
            msg = (
                f"Metric {self._monitor} improved by {abs(self.best_score - current):.3f} >="
                f" min_delta = {abs(self._min_delta)}. New best score: {current:.3f}"
            )
        else:
            msg = f"Metric {self._monitor} improved. New best score: {current:.3f}"
        return msg

    @staticmethod
    def _log_info(trainer: "reax.Trainer", message: str, log_rank_zero_only: bool) -> None:
        """Log info."""
        rank = trainer.process_index if trainer.process_count > 1 else None
        message = rank_zero.rank_prefixed_message(message, rank)
        if rank is None or not log_rank_zero_only or rank == 0:
            _LOGGER.info(message)



import abc
from typing import TypeVar

import jax
import jax.numpy as jnp
from typing_extensions import override

from reax.utils import arrays

from . import utils
from ._metric import Metric

__all__ = ("Average", "Std", "Min", "Max", "Unique", "NumUnique", "Sum")


class Aggregation(Metric[jax.Array], abc.ABC):
    """Interface that defines an aggregation metric, i.e. one that take raw array-like data and an
    optional mask.
    """

    Self = TypeVar("Self", bound="Aggregation")

    @abc.abstractmethod
    def create(  # pylint: disable=arguments-differ
        self,
        values: jax.typing.ArrayLike,
        mask: jax.typing.ArrayLike | None = None,
    ) -> Self:
        """Create the metric from data."""

    @abc.abstractmethod
    def update(  # pylint: disable=arguments-differ
        self,
        values: jax.typing.ArrayLike,
        mask: jax.typing.ArrayLike | None = None,
    ) -> Self:
        """Update this metric and return an updated instance."""


class Sum(utils.WithAccumulator, Aggregation):
    reduce_fn = jnp.sum


class Min(utils.WithAccumulator, Aggregation):
    reduce_fn = jnp.min


class Max(utils.WithAccumulator, Aggregation):
    reduce_fn = jnp.max


class Unique(utils.WithAccumulator, Aggregation):
    """Get the set of unique values.

    warning: this cannot be used with JAX jit because it relies on dynamically-sized arrays
    """

    @staticmethod
    def reduce_fn(values, where=None):
        """Reduce fn."""
        np_ = arrays.infer_backend([values, where])
        return np_.unique(values[where] if where is not None else values)


class NumUnique(utils.WithAccumulator, Aggregation):
    """Count the number of unique values.

    Warning:
        this cannot be used with JAX jit because it relies on dynamically-sized arrays.....
    """

    @staticmethod
    def reduce_fn(values, where=None):
        """Reduce fn."""
        np_ = arrays.infer_backend([values, where])
        return np_.unique(values[where] if where is not None else values)

    def compute(self) -> jax.Array:
        """Compute function."""
        np_ = arrays.infer_backend(self.accumulator)
        return np_.asarray(self.accumulator.size)


class Average(utils.WithAccumulatorAndCount, Aggregation):
    """Compute an average"""

    reduce_fn = jnp.sum


class Std(Aggregation):
    """Calculate standard deviation."""

    total: jax.Array
    sum_of_squares: jax.Array
    count: jax.Array

    def __init__(
        self, total: jax.Array = None, sum_of_squares: jax.Array = None, count: jax.Array = None
    ):
        super().__init__()
        # State
        self.total = total or jnp.array(0, jnp.float32)
        self.sum_of_squares = sum_of_squares or jnp.array(0, jnp.float32)
        self.count = count or jnp.array(0, jnp.int32)

    @override
    @classmethod
    def empty(cls) -> "Metric[jax.Array]":
        return cls()

    @override
    @classmethod
    def create(  # pylint: disable=arguments-differ
        cls,
        values: jax.typing.ArrayLike,
        mask: jax.typing.ArrayLike | None = None,
    ) -> "Std":
        """Create the metric from data."""
        if values.ndim == 0:
            values = values[None]
        if mask is None:
            mask = jnp.ones(values.shape[0], dtype=jnp.bool)

        mask, num_elements = utils.prepare_mask(values, mask, return_count=True)
        return cls(
            total=values.sum(),
            sum_of_squares=jnp.where(mask, values**2, jnp.zeros_like(values)).sum(),
            count=num_elements,
        )

    def update(
        # pylint: disable=arguments-differ
        self,
        values: jax.Array,
        mask: jax.Array | None = None,
    ) -> "Std":
        """Update function."""
        if values.ndim == 0:
            values = values[None]
        if mask is None:
            mask = jnp.ones(values.shape[0], dtype=jnp.bool)

        mask, num_elements = utils.prepare_mask(values, mask, return_count=True)
        return type(self)(
            total=self.total + values.sum(),
            sum_of_squares=self.sum_of_squares
            + jnp.where(mask, values**2, jnp.zeros_like(values)).sum(),
            count=self.count + num_elements,
        )

    def merge(self, other: "Std") -> "Std":
        """Merge function."""
        return type(self)(
            total=self.total + other.total,
            sum_of_squares=self.sum_of_squares + other.sum_of_squares,
            count=self.count + other.count,
        )

    def compute(self) -> jax.Array:
        """Compute function."""
        # var(X) = 1/N \sum_i (x_i - mean)^2
        #        = 1/N \sum_i (x_i^2 - 2 x_i mean + mean^2)
        #        = 1/N ( \sum_i x_i^2 - 2 mean \sum_i x_i + N * mean^2 )
        #        = 1/N ( \sum_i x_i^2 - 2 mean N mean + N * mean^2 )
        #        = 1/N ( \sum_i x_i^2 - N * mean^2 )
        #        = \sum_i x_i^2 / N - mean^2
        mean = self.total / self.count
        variance = self.sum_of_squares / self.count - mean**2
        # Mathematically, variance can never be negative but in reality we may run
        # into such issues for numerical reasons.
        variance = jnp.clip(variance, min=0.0)
        return variance**0.5



from typing import TYPE_CHECKING, Any, Final

import beartype
from flax import nnx
import jaxtyping as jt
from lightning_utilities.core import overrides
from typing_extensions import override

from . import common, stages, train, validation
from .. import data, exceptions, modules

if TYPE_CHECKING:
    import reax

__all__ = "FitEpoch", "Fit"


class FitEpoch(train.Train):
    @jt.jaxtyped(typechecker=beartype.beartype)
    def __init__(
        self,
        module: "reax.Module",
        datamanager: "reax.data.DataSourceManager",
        optimizers: list["reax.Optimizer"],
        engine: "reax.Engine",
        *,
        rngs: nnx.Rngs = None,
        fast_dev_run: bool | int = False,
        min_updates: int | None = None,
        max_updates: int | float | None = None,
        limit_train_batches: int | float | None = 1.0,
        accumulate_grad_batches: int = 1,
        limit_val_batches: int | float | None = 1.0,
        val_check_interval: int | float | None = 1.0,
        check_val_every_n_epoch: int = 1,
        stopper: common.Stopper | None = None,
    ):
        """Init function."""
        if fast_dev_run:
            if isinstance(fast_dev_run, int):
                if fast_dev_run < 0:
                    raise exceptions.MisconfigurationException("`fast_dev_run` should be >= 0")
                if fast_dev_run == 1:
                    fast_dev_run = True

            num_batches = 1 if fast_dev_run is True else fast_dev_run
            limit_train_batches = num_batches
            limit_val_batches = num_batches
            val_check_interval = 1.0
            check_val_every_n_epoch = 1

        super().__init__(
            module,
            datamanager,
            engine,
            optimizers,
            rngs=rngs,
            fast_dev_run=fast_dev_run,
            min_updates=min_updates,
            max_updates=max_updates,
            limit_batches=limit_train_batches,
            accumulate_grad_batches=accumulate_grad_batches,
            stopper=stopper,
        )
        # Params
        self._val_check_interval: Final[int | float] = val_check_interval
        self._check_val_every_n_epoch: Final[int] = check_val_every_n_epoch

        # State
        if (
            not self._datamanager.has_dataloader("val")
            or limit_val_batches == 0.0
            or not overrides.is_overridden("validation_step", module, modules.Module)
        ):
            # No validation
            self._validate = None
        else:
            self._validate = validation.Validate(
                module,
                datamanager,
                engine,
                fast_dev_run=fast_dev_run,
                limit_batches=limit_val_batches,
            )
        self._val_check_batch = None

    @property
    def train_dataloader(self) -> "reax.DataLoader":
        """Train dataloader."""
        return self.dataloader

    @property
    def val_dataloader(self) -> "reax.DataLoader | None":
        """Train dataloader."""
        if self._validate is None:
            return None

        return self._validate.dataloader

    @property
    def val_check_interval(self) -> int | float | None:
        """Val check interval."""
        return self._val_check_interval

    @property
    def check_val_every_n_epoch(self) -> int | None:
        """Check val every n epoch."""
        return self._check_val_every_n_epoch

    @property
    def limit_train_batches(self) -> int | float | None:
        """The limit to the number of batches in a train epoch"""
        return self._limit_batches

    @property
    def limit_val_batches(self) -> int | float | None:
        """The limit to the number of batches in a validation epoch"""
        if self._validate is None:
            return None

        return self._validate.limit_batches

    @property
    def validate(self) -> validation.Validate | None:
        """Validate function."""
        return self._validate

    @override
    def _on_starting(self):
        """On starting."""
        super()._on_starting()
        # The superclass will have prepared the dataloader so we can now configure when to check
        # validation (if at all)
        self._val_check_batch = self._setup_val_check_batch_(
            self._val_check_interval,
            self.max_batches,
            self._check_val_every_n_epoch,
            self.train_dataloader,
        )

        # Only the root stage does setup as this only needs to be done once per stage tree
        if self.is_root and self._module is not None:
            self._module.setup(self, next(iter(self.train_dataloader)))
            params = self._engine.to_device(self._module.parameters())
            self._module.set_parameters(params)

    @override
    def _on_iteration_finished(self, outputs: Any, /) -> None:
        """On iteration finished."""
        super()._on_iteration_finished(outputs)

        # We've finished the train iteration, so check if we should do a validation
        # if (
        #     isinstance(self._val_check_interval, int)
        #     and self.iteration % self._val_check_interval == 0
        # ):
        #     self._run_child(self._validate)

        if self._should_check_val():
            self._run_child(self._validate)

    @override
    def _on_stopping(self) -> None:
        """On stopping."""
        # if (
        #     self._validate is not None
        #     and self._check_val_every_n_epoch is not None
        #     and (self.epoch + 1) % self._check_val_every_n_epoch == 0
        # ):
        #     self._run_child(self._validate)
        super()._on_stopping()

    @override
    def log(
        self,
        name: str,
        value: "jt.ArrayLike | reax.typing.MetricInstance",
        batch_size: int | None = None,
        prog_bar: bool = False,
        logger: bool = False,
        on_step=False,
        on_epoch=True,
    ) -> None:
        """Log function."""
        if self._child is not None:
            self._child.log(name, value, batch_size, prog_bar, logger, on_step, on_epoch)
        else:
            super().log(name, value, batch_size, prog_bar, logger, on_step, on_epoch)

    def _should_check_val_epoch(self) -> bool:
        """Should check val epoch."""
        return self.validate and (
            self._check_val_every_n_epoch is None
            or (self.epoch + 1) % self._check_val_every_n_epoch == 0
        )

    def _should_check_val(self):
        """Decide if we should run validation."""
        if not self._should_check_val_epoch():
            return False

        # val_check_batch is inf for iterable datasets with no length defined
        is_infinite_dataset = self._val_check_batch == float("inf")
        # is_last_batch = self.batch_progress.is_last_batch
        # if is_last_batch and (
        #     is_infinite_dataset or isinstance(data_fetcher, _DataLoaderIterDataFetcher)
        # ):
        #     return True

        if self.should_stop:
            # allow validation if requesting to stop early through `Trainer.should_stop`
            # (e.g. by early stopping) and when the loop allows to stop (min_epochs/steps met)
            return True

        # TODO: let training/eval loop handle logic around limit_*_batches and val_check_batch
        # is_val_check_batch = is_last_batch
        is_val_check_batch = False
        if isinstance(self.limit_train_batches, int) and is_infinite_dataset:
            is_val_check_batch = (self.batch_idx + 1) % self.limit_train_batches == 0
        elif self._val_check_batch != float("inf"):
            # if `check_val_every_n_epoch is `None`, run a validation loop every n training batches
            # else condition it based on the batch_idx of the current epoch
            current_iteration = (
                self.total_batch_idx if self.check_val_every_n_epoch is None else self.batch_idx
            )
            is_val_check_batch = (current_iteration) % self._val_check_batch == 0

        return is_val_check_batch

    @staticmethod
    def _setup_val_check_batch_(
        val_check_interval: int | float,
        max_batches: int | float,
        check_val_every_n_epoch: int | None,
        dataloader: "reax.DataLoader",
    ) -> int | float | None:
        """Setup val check batch."""
        if max_batches == 0:
            return None

        if isinstance(val_check_interval, int):
            val_check_batch = val_check_interval
            if val_check_batch > max_batches and check_val_every_n_epoch is not None:
                raise ValueError(
                    f" `val_check_interval` ({val_check_interval}) must be less than or equal"
                    f" to the number of the training batches ({max_batches}). "
                    "If you want to disable validation set `limit_val_batches` to 0.0 instead."
                    "If you want to validate based on the total training batches, set "
                    "`check_val_every_n_epoch=None`."
                )
        else:
            dataloader_size = data.sized_len(dataloader)
            has_len_all_ranks_ = dataloader_size is not None
            if not has_len_all_ranks_:
                if val_check_interval == 1.0:
                    val_check_batch = float("inf")
                else:
                    raise exceptions.MisconfigurationException(
                        "When using an IterableDataset for `train_dataloader`,"
                        " `Trainer(val_check_interval)` must be `1.0` or an int. An int k specifies"
                        " checking validation every k training batches."
                    )
            else:
                val_check_batch = int(max_batches * val_check_interval)
                val_check_batch = max(1, val_check_batch)

        # if loggers and max_batches < log_every_n_steps and not fast_dev_run:
        #     rank_zero.rank_zero_warn(
        #         f"The number of training batches ({max_batches}) is smaller than the logging "
        #         f"interval. Trainer(log_every_n_steps={log_every_n_steps}). Set a lower value "
        #         f"for log_every_n_steps if you want to see logs for the training epoch.",
        #         category=PossibleUserWarning,
        #     )

        return val_check_batch


class Fit(stages.Stage):
    def __init__(
        self,
        module: "reax.Module",
        datamanager: "reax.data.DataSourceManager",
        optimizers: list["reax.Optimizer"],
        engine: "reax.Engine",
        *,
        rngs: nnx.Rngs = None,
        fast_dev_run: bool | int = False,
        max_epochs: int | None = None,
        min_epochs: int = 0,
        min_updates: int = 0,
        max_updates: int | float | None = None,
        limit_train_batches: int | float | None = 1.0,
        accumulate_grad_batches: int = 1,
        limit_val_batches: int | float | None = 1.0,
        val_check_interval: int | float | None = 1.0,
        check_val_every_n_epoch: int = 1,
        num_sanity_val_steps: int | None = 2,
        reload_dataloaders_every_n_epochs: int = 0,
    ):
        """Init function."""
        if (
            not isinstance(reload_dataloaders_every_n_epochs, int)
            or reload_dataloaders_every_n_epochs < 0
        ):
            raise exceptions.MisconfigurationException(
                f"`reload_dataloaders_every_n_epochs` should be an int >= 0, got "
                f"{reload_dataloaders_every_n_epochs}."
            )

        if fast_dev_run:
            max_epochs = 1
            num_sanity_val_steps = 0

        super().__init__(
            "fit",
            module,
            engine,
            rngs=rngs,
            datamanager=datamanager,
            max_iters=max_epochs,
            min_iters=min_epochs,
        )

        # Params
        self._num_sanity_val_steps: Final[int | None] = num_sanity_val_steps
        self._reload_dataloaders_every_n_epochs: Final[int] = reload_dataloaders_every_n_epochs

        # State
        self._fit_epoch = FitEpoch(
            module,
            datamanager,
            optimizers,
            engine,
            rngs=rngs,
            fast_dev_run=fast_dev_run,
            min_updates=min_updates,
            max_updates=max_updates,
            limit_train_batches=limit_train_batches,
            accumulate_grad_batches=accumulate_grad_batches,
            limit_val_batches=limit_val_batches,
            val_check_interval=val_check_interval,
            check_val_every_n_epoch=check_val_every_n_epoch,
            stopper=self._stopper,
        )
        self._sanity_check: validation.Validate | None = None
        if self._fit_epoch.validate and num_sanity_val_steps:
            if num_sanity_val_steps is not None and limit_val_batches is not None:
                batches_limit = min(num_sanity_val_steps, limit_val_batches)
            elif num_sanity_val_steps is not None:
                batches_limit = num_sanity_val_steps
            else:
                batches_limit = limit_val_batches

            self._sanity_check = validation.Validate(
                module,
                datamanager,
                engine,
                fast_dev_run=fast_dev_run,
                limit_batches=batches_limit,
                name="sanity_check",
                enable_checkpointing=False,
            )

    @property
    def epoch(self) -> int:
        return self.iteration

    @property
    def fast_dev_run(self) -> bool | int:
        return self._fit_epoch.fast_dev_run

    @property
    def sanity_checking(self) -> bool:
        """`True` if currently sanity checking, `False` otherwise"""
        return self._sanity_check is not None and self._child is self._sanity_check

    @property
    def max_epochs(self) -> int | None:
        return self.max_iters

    @property
    def updates(self) -> int:
        """Updates function."""
        return self._fit_epoch.updates

    @property
    def enable_validation(self) -> bool:
        """Returns `True` of validation is enabled, `False` otherwise"""
        return self._fit_epoch.validate is not None

    @property
    def validate(self) -> validation.Validate | None:
        """Validate function."""
        return self._fit_epoch.validate

    @property
    def limit_train_batches(self) -> int | float | None:
        """The limit on the number of training batches per epoch"""
        return self._fit_epoch.limit_train_batches

    @property
    def num_training_batches(self) -> int | float | None:
        return self._fit_epoch.num_training_batches

    @property
    def limit_val_batches(self) -> int | float | None:
        """The limit on the number of training batches per epoch"""
        return self._fit_epoch.limit_val_batches

    @property
    def num_val_batches(self) -> int | float | None:
        if not self.enable_validation:
            return None

        return self._fit_epoch.validate.num_batches

    @property
    def num_sanity_val_steps(self) -> int | None:
        return self._num_sanity_val_steps

    @property
    def val_check_interval(self):
        """Val check interval."""
        return self._fit_epoch.val_check_interval

    @property
    def check_val_every_n_epoch(self) -> int | None:
        """Check val every n epoch."""
        return self._fit_epoch.check_val_every_n_epoch

    @property
    def train_dataloader(self) -> "reax.DataLoader":
        """Train dataloader."""
        return self._fit_epoch.train_dataloader

    @property
    def train_dataloaders(self) -> "reax.DataLoader":
        return self.train_dataloader

    @property
    def val_dataloader(self) -> "reax.DataLoader":
        """Train dataloader."""
        return self._fit_epoch.val_dataloader

    @property
    def val_dataloaders(self) -> "reax.DataLoader":
        return self.val_dataloader

    @override
    def log(
        self,
        name: str,
        value,
        batch_size: int | None = None,
        prog_bar: bool = False,
        logger: bool = False,
        on_step=False,
        on_epoch=True,
    ) -> None:
        """Log function."""
        if self._child is None:
            raise RuntimeError(
                "Fitting is not running a train, validation or sanity check epoch, "
                "so cannot currently log anything"
            )

        self._child.log(name, value, batch_size, prog_bar, logger, on_step, on_epoch)

    @override
    def _on_iteration_starting(self):
        super()._on_iteration_starting()
        if (
            self._reload_dataloaders_every_n_epochs
            and self.epoch % self._reload_dataloaders_every_n_epochs == 0
        ):
            self._datamanager.reset()

    @override
    def _step(self) -> Any:
        """Step function."""
        if self.iteration == 0 and self._sanity_check:
            self._run_child(self._sanity_check)

        self._run_child(self._fit_epoch)



from collections.abc import Iterable, Iterator
from typing import TYPE_CHECKING, Optional, TypeVar

import beartype
import jax
import jaxtyping as jt
import numpy as np
from typing_extensions import override

from . import _types, collate, fetchers, samplers

if TYPE_CHECKING:
    import reax

__all__ = "ArrayLoader", "CachingLoader", "ReaxDataLoader", "FetcherDataLoader", "DeviceDataLoader"


T = TypeVar("T")
_U = TypeVar("_U")
_T_co = TypeVar("_T_co", covariant=True)
ArrayOrArrayTuple = jt.ArrayLike | tuple[jt.ArrayLike, ...]


def _single_or_value(value: tuple[T, ...], to_test=None) -> T | tuple[T, ...]:
    """Single or value."""
    if to_test is None:
        to_test = value
    if len(to_test) > 1:
        return tuple(value)
    return value[0]


class FetcherDataLoader(_types.DataLoader[_T_co, _U]):
    """A data loader that uses a fetcher to get samples"""

    @jt.jaxtyped(typechecker=beartype.beartype)
    def __init__(
        self,
        dataset: "reax.data.Dataset[_T_co]",
        sampler: _types.Sampler | Iterable,
        fetcher: "fetchers._BaseFetcher[_T_co, _U]",
    ):
        self._dataset = dataset
        self._sampler = sampler
        self._fetcher = fetcher

    @override
    @property
    def dataset(self) -> "reax.data.Dataset[_T_co]":
        return self._dataset

    @override
    @property
    def sampler(self):
        return self._sampler

    def __iter__(self) -> Iterator[_U]:
        """Iter function."""
        try:
            for indices in self.sampler:
                yield self._fetcher.fetch(indices)
        except StopIteration:
            pass

    def with_new_sampler(self, sampler: "reax.data.Sampler") -> "FetcherDataLoader[_T_co, _U]":
        return FetcherDataLoader(dataset=self._dataset, sampler=sampler, fetcher=self._fetcher)


class ReaxDataLoader(FetcherDataLoader[_T_co, _U]):
    """A general-purpose data loader provided by REAX that is suitable for most situations."""

    @jt.jaxtyped(typechecker=beartype.beartype)
    def __init__(
        self,
        dataset: "reax.data.Dataset[_T_co]",
        batch_size: int = 1,
        shuffle: bool = False,
        sampler=None,
        collate_fn: _types.CollateFn | None = None,
    ):
        sampler = samplers.create_sampler(
            dataset, batch_size=batch_size, shuffle=shuffle, sampler=sampler
        )

        if collate_fn is None:
            collate_fn = collate.get_default_collator().collate

        fetcher = fetchers.create_fetcher(dataset, collate_fn=collate_fn)

        super().__init__(dataset, sampler, fetcher)


class ArrayLoader(_types.DataLoader[ArrayOrArrayTuple, ArrayOrArrayTuple]):
    """A more efficient loader for array datasets that uses .take() to batch over slices of an
    array dataset.
    """

    @jt.jaxtyped(typechecker=beartype.beartype)
    def __init__(
        self,
        arrays: ArrayOrArrayTuple,
        batch_size: int = 1,
        shuffle=False,
        sampler: "Optional[reax.data.Sampler]" = None,
    ):
        # Params
        self._batch_size: int = batch_size
        self._shuffle: bool = shuffle

        if isinstance(arrays, tuple):
            if not all(arrays[0].shape[0] == array.shape[0] for array in arrays):
                raise ValueError("Size mismatch between tensors")
            first_array = arrays[0]
        else:
            if not isinstance(arrays, (jax.Array, np.ndarray)):
                raise TypeError(f"Expected array or tuple of arrays, got {type(arrays).__name__}")
            first_array = arrays

        # State
        self._arrays = arrays
        self._sampler = samplers.create_sampler(
            first_array, batch_size=batch_size, shuffle=shuffle, sampler=sampler
        )

    def __iter__(self) -> Iterator[ArrayOrArrayTuple]:
        """Iter function."""
        for idx in self._sampler:
            idx = np.asarray(idx)
            if isinstance(self._arrays, tuple):
                yield tuple(array.take(idx, axis=0) for array in self._arrays)
            else:
                yield self._arrays.take(idx, axis=0)

    def __len__(self) -> int:
        """Len function."""
        return len(self._sampler)

    @property
    def batch_size(self) -> int:
        return self._batch_size

    @property
    def shuffle(self) -> bool:
        return self._shuffle

    @property
    def dataset(self) -> ArrayOrArrayTuple:
        return self._arrays

    @property
    def sampler(self):
        return self._sampler

    def first(self) -> ArrayOrArrayTuple:
        """First function."""
        return next(iter(self))

    def with_new_sampler(self, sampler: "reax.data.Sampler") -> "ArrayLoader":
        return ArrayLoader(
            arrays=self.dataset, batch_size=self._batch_size, shuffle=self._shuffle, sampler=sampler
        )


class CachingLoader(_types.DataLoader[_T_co, _U]):
    """Caching loader is useful, for example, if you don't want to shuffle data every time but at
    some interval defined by ``reset_every``.

    This means you need to have enough memory to accommodate all the data.
    """

    def __init__(self, loader: "reax.data.DataLoader[_T_co]", reset_every: int):
        self._loader = loader
        self._reset_every = reset_every
        self._time_since_reset = 0
        self._cache = None

    def __iter__(self) -> Iterator[_U]:
        """Iter function."""
        if self._cache:
            yield from self._cache
        else:
            # Have to pull from the loader
            cache = []
            for entry in self._loader:
                yield entry
                cache.append(entry)
            self._cache = cache

        self._time_since_reset += 1
        # Check if we should clear the cache
        if self._time_since_reset == self._reset_every:
            self._cache = []
            self._time_since_reset = 0

    @override
    @property
    def dataset(self) -> "reax.data.Dataset[_T_co]":
        return self._loader.dataset

    @override
    @property
    def sampler(self):
        return self._loader.sampler

    @override
    def with_new_sampler(self, sampler: "reax.data.Sampler") -> "CachingLoader[_T_co, _U]":
        return CachingLoader(self._loader.with_new_sampler(sampler), reset_every=self._reset_every)


class DeviceDataLoader(_types.DataLoader[_T_co, _U]):
    """A loader that wraps an existing dataloader the puts data onto the specified device."""

    def __init__(self, loader: "reax.DataLoader[_T_co, _U]", device: jax.Device):
        self._loader = loader
        self._device = device

    def __iter__(self) -> Iterator[_U]:
        for data in self._loader:
            yield jax.device_put(data, device=self._device)

    @property
    def parent(self) -> "reax.data.DataLoader[_T_co, _U]":
        """Access the original loader that this one wraps"""
        return self._loader

    @override
    @property
    def dataset(self) -> "reax.data.Dataset[_T_co, _U]":
        return self._loader.dataset

    @override
    @property
    def sampler(self) -> "reax.data.Sampler":
        return self._loader.sampler

    @override
    def with_new_sampler(self, sampler: "reax.data.Sampler") -> "DataLoader[_T_co, U]":
        return DeviceDataLoader(self._loader.with_new_sampler(sampler), device=self._device)



import jax
import jax.numpy as jnp
from typing_extensions import override

from . import _metric, aggregation

__all__ = "MeanSquaredError", "RootMeanSquareError", "MeanAbsoluteError", "LeastSquaresEstimate"

MeanSquaredError = aggregation.Average.from_fun(
    lambda values, targets, mask=None: (jnp.square(values - targets), mask),
    name="SquaredError",
)


MeanAbsoluteError = aggregation.Average.from_fun(
    lambda values, targets, mask=None: (jnp.abs(values - targets), mask),
    name="AbsoluteError",
)


class RootMeanSquareError(_metric.Metric):
    mse: MeanSquaredError

    def __init__(self, mse: MeanSquaredError = None):
        """Init function."""
        super().__init__()
        # State
        self.mse = mse

    @property
    def is_empty(self) -> bool:
        """Is empty."""
        return self.mse is None

    @classmethod
    @override
    def create(  # pylint: disable=arguments-differ
        cls, values: jax.Array, targets: jax.Array, mask: jax.Array | None = None, /
    ) -> "RootMeanSquareError":
        """Create function."""
        return cls(mse=MeanSquaredError().create(values, targets, mask))

    def update(  # pylint: disable=arguments-differ
        self,
        values: jax.Array,
        targets: jax.Array,
        mask: jax.Array | None = None,
    ) -> "RootMeanSquareError":
        """Update function."""
        if self.is_empty:
            return self.create(values, targets, mask)

        return type(self)(mse=self.mse.update(values, targets, mask=mask))

    def merge(self, other: "RootMeanSquareError") -> "RootMeanSquareError":
        """Merge function."""
        if self.is_empty:
            return other
        if other.is_empty:
            return self

        return type(self)(mse=self.mse.merge(other.mse))

    def compute(self) -> jax.Array:
        """Compute function."""
        return jnp.sqrt(self.mse.compute())


class LeastSquaresEstimate(_metric.Metric[jax.Array]):
    values: jax.Array | None
    targets: jax.Array | None

    def __init__(self, values: jax.Array = None, targets: jax.Array = None):
        """Init function."""
        super().__init__()
        self.values = values
        self.targets = targets

    @property
    def is_empty(self) -> bool:
        """Is empty."""
        return self.values is None

    @classmethod
    @override
    def empty(cls) -> "LeastSquaresEstimate":
        return cls()

    @classmethod
    @override
    def create(  # pylint: disable=arguments-differ
        cls, inputs: jax.Array, outputs: jax.Array
    ) -> "LeastSquaresEstimate":
        """Create function."""
        return cls(inputs, outputs)

    @override
    def update(  # pylint: disable=arguments-differ
        self, inputs: jax.Array, outputs: jax.Array
    ) -> "LeastSquaresEstimate":
        """Update function."""
        if self.is_empty:
            return self.create(inputs, outputs)

        return type(self)(
            values=jnp.concatenate((self.values, inputs)),
            targets=jnp.concatenate((self.targets, outputs)),
        )

    @override
    def merge(self, other: "LeastSquaresEstimate") -> "LeastSquaresEstimate":
        """Merge function."""
        if self.is_empty:
            return other
        if other.is_empty:
            return self

        return LeastSquaresEstimate(
            values=jnp.concatenate((self.values, other.values)),
            targets=jnp.concatenate((self.targets, other.targets)),
        )

    @override
    def compute(self) -> jax.Array:
        """Compute function."""
        return jnp.linalg.lstsq(self.values, self.targets)[0]



from typing import TYPE_CHECKING, Any, TypeVar

import jax
import jaxtyping as jt
from typing_extensions import override

from . import _strategies

if TYPE_CHECKING:
    import reax

__all__ = ("SingleDevice",)

_OutT = TypeVar("_OutT")


class SingleDevice(_strategies.Strategy):
    """Strategy for a single device e.g. a single GPU."""

    def __init__(self, platform: str):
        self._device = jax.Device = (
            jax.devices()[0] if platform == "auto" else jax.devices(platform)[0]
        )

    @property
    def device(self) -> jax.Device:
        return self._device

    @override
    def to_device(self, value: Any) -> Any:
        """To device."""
        return jax.device_put(value, self._device)

    @override
    def from_device(self, value: Any) -> Any:
        """From device."""
        return jax.device_get(value)

    @property
    @override
    def is_global_zero(self) -> bool:
        """Is global zero."""
        return True

    @override
    def broadcast(self, obj: jt.PyTreeDef, src: int = 0) -> Any:
        """Broadcast function."""
        return obj

    @override
    def all_gather(self, obj: jt.PyTreeDef) -> Any:
        return obj

    @override
    def all_reduce(self, obj: jt.PyTree, reduce_op: str = "mean") -> jt.PyTree:
        """Reduces a tensor from several distributed processes to one aggregated tensor.

        Args:
            obj: the pytree to sync and reduce
            reduce_op: the reduction operation. Defaults to 'mean'/'avg'.
                Can also be a string 'sum' to calculate the sum during reduction.

        Return:
            reduced value
        """
        return obj

    @override
    def barrier(self, name: str | None = None) -> None:
        """Synchronizes all processes which blocks processes until the whole group enters this
        function.

        Args:
            name: an optional name to pass into barrier.
        """
        return None  # Nothing to do in a single device

    @override
    def compute(self, metric: "reax.typing.MetricInstance[_OutT]") -> _OutT:
        return metric.compute()



# Copyright (C) 2024  Martin Uhrin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Most of this file is covered by the following license.  To find what has been modified you
# can perform a diff with the file at:
# https://github.com/Lightning-AI/pytorch-lightning/blob/0324a20f00235c7a10a235a44326811ba42b6ae4/src/lightning/pytorch/callbacks/progress/progress_bar.py
#
# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Any, Optional

from lightning_utilities.core import rank_zero
from typing_extensions import override

from reax import hooks

if TYPE_CHECKING:
    import reax

__all__ = ("ProgressBar",)


class ProgressBar(hooks.TrainerListener):
    r"""The base class for progress bars in REAX. It is a :class:`~reax.listeners.TrainerListener`
    that keeps track of the batch progress in the :class:`~reax.trainer.Trainer`. You should
    implement your highly custom progress bars with this as the base class.

    Example::

        class ReaxProgressBar(ProgressBar):

            def __init__(self):
                super().__init__()  # don't forget this :)
                self.enable = True

            def disable(self):
                self.enable = False

            def on_train_batch_end(self, trainer, stage, outputs, batch, batch_idx):
                # don't forget this :)
                super().on_train_batch_end(trainer, stage, outputs, batch, batch_idx)
                percent = (batch_idx / self.total_train_batches) * 100
                sys.stdout.flush()
                sys.stdout.write(f'{percent:.01f} percent complete \r')

        bar = ReaxProgressBar()
        trainer = Trainer(callbacks=[bar])

    """

    def __init__(self) -> None:
        self._trainer: "Optional[reax.Trainer]" = None
        self._current_eval_dataloader_idx: int | None = None

    @property
    def sanity_check_description(self) -> str:
        return "Sanity Checking"

    @property
    def train_description(self) -> str:
        return "Training"

    @property
    def validation_description(self) -> str:
        return "Validation"

    @property
    def test_description(self) -> str:
        return "Testing"

    @property
    def predict_description(self) -> str:
        return "Predicting"

    def disable(self) -> None:
        """You should provide a way to disable the progress bar."""
        raise NotImplementedError

    def enable(self) -> None:
        """You should provide a way to enable the progress bar.

        The :class:`~lightning.pytorch.trainer.trainer.Trainer` will call this in e.g. pre-training
        routines like the :ref:`learning rate finder
        <advanced/training_tricks:Learning Rate Finder>`. to temporarily enable and disable the
        training progress bar.

        """
        raise NotImplementedError

    def print(self, *args: Any, **kwargs: Any) -> None:
        """You should provide a way to print without breaking the progress bar."""
        print(*args, **kwargs)

    @override
    def setup(self, trainer: "reax.Trainer", _stage: "reax.Stage", /) -> None:
        self._trainer = trainer
        if not trainer.is_global_zero:
            self.disable()

    def get_metrics(
        self, trainer: "reax.Trainer", *_
    ) -> dict[str, int | str | float | dict[str, float]]:
        r"""Combines progress bar metrics collected from the trainer with standard metrics from
        get_standard_metrics. Implement this to override the items displayed in the progress bar.

        Here is an example of how to override the defaults:

        .. code-block:: python

            def get_metrics(self, trainer, model):
                # don't show the version number
                items = super().get_metrics(trainer, model)
                items.pop("v_num", None)
                return items

        Return:
            Dictionary with the items to be displayed in the progress bar.

        """
        standard_metrics = get_standard_metrics(trainer)
        pbar_metrics = trainer.progress_bar_metrics
        duplicates = list(standard_metrics.keys() & pbar_metrics.keys())
        if duplicates:
            rank_zero.rank_zero_warn(
                f"The progress bar already tracks a metric with the name(s) "
                f"'{', '.join(duplicates)}' and `self.log('{duplicates[0]}', ..., prog_bar=True)` "
                f"will overwrite this value. If this is undesired, change the name or override "
                f"`get_metrics()` in the progress bar callback."
            )

        return {**standard_metrics, **pbar_metrics}


def get_standard_metrics(trainer: "reax.Trainer") -> dict[str, int | str]:
    r"""Returns the standard metrics displayed in the progress bar. Currently, it only includes the
    version of the experiment when using a logger.

    .. code-block::

        Epoch 1:   4%|▎         | 40/1095 [00:03<01:37, 10.84it/s, v_num=10]

    Return:
        Dictionary with the standard metrics to be displayed in the progress bar.

    """
    items_dict: dict[str, int | str] = {}
    if trainer.loggers:

        if (version := _version(trainer.loggers)) not in ("", None):
            if isinstance(version, str):
                # show last 4 places of long version strings
                version = version[-4:]
            items_dict["v_num"] = version

    return items_dict


def _version(loggers: list[Any], separator: str = "_") -> int | str:
    if len(loggers) == 1:
        return loggers[0].version
    # Concatenate versions together, removing duplicates and preserving order
    return separator.join(dict.fromkeys(str(logger.version) for logger in loggers))



import collections.abc
from collections.abc import Iterable, Iterator
from typing import Any

from flax import linen
import jax
import jax.numpy as jnp
import numpy as np
import optax
from typing_extensions import override

import reax
from reax import data, modules

__all__ = "BoringModel", "RandomDataset", "RandomIterableDataset"


class BoringModel(modules.Module):
    """Testing REAX Module.

    Use as follows:
    - subclass
    - modify the behavior for what you want

    Warning:
        This is meant for testing/debugging and is experimental.

    Example::

        class TestModel(BoringModel):
            def training_step(self, ...):
                ...  # do your own thing
    """

    def __init__(self) -> None:
        """Init function."""
        super().__init__()
        self.layer = linen.Dense(2)

    @override
    def configure_model(self, _stage, batch: Any, /) -> None:
        """Configure our model."""
        if self._parameters is None:
            params = self.layer.init(self.rngs(), batch[0])
            self.set_parameters(params)

    def forward(self, x: jax.Array) -> jax.Array:
        """Forward function."""
        return self.layer.apply(self._parameters, x)

    def __call__(self, *args, **kwargs):
        """Call function."""
        return self.forward(*args, **kwargs)

    @staticmethod
    def loss(preds: jax.Array, labels: jax.Array | None = None) -> jax.Array:
        """Loss function."""
        if labels is None:
            labels = jnp.ones_like(preds)
        # An arbitrary loss to have a loss that updates the model weights during `Trainer.fit` calls
        return optax.losses.squared_error(preds, labels).mean()

    @staticmethod
    def step(model: linen.Module, parameters, batch: Any) -> jax.Array:
        """Step function."""
        output = model.apply(parameters, batch)
        return BoringModel.loss(output)

    @override
    def training_step(self, batch: Any, _batch_idx: int, /) -> Any:
        """Training step."""
        loss, grad = jax.value_and_grad(self.step, argnums=1)(self.layer, self._parameters, batch)
        return {"loss": loss, "grad": grad}

    @override
    def validation_step(self, batch: Any, _batch_idx: int, /) -> Any:
        """Validation step."""
        return {"x": self.step(self.layer, self._parameters, batch)}

    @override
    def test_step(self, batch: Any, _batch_idx: int, /) -> Any:
        """Test step."""
        return {"y": self.step(self.layer, self._parameters, batch)}

    @override
    def predict_step(self, batch: Any, _batch_idx: int, /) -> Any:
        """Predict step."""
        return self.forward(batch)

    def configure_optimizers(self) -> tuple:
        """Configure optimizers."""
        schedule = optax.exponential_decay(init_value=0.1, transition_steps=1, decay_rate=0.1)
        optimizer = optax.sgd(learning_rate=schedule)
        return optimizer, optimizer.init(self._parameters)

    def train_dataloader(self) -> data.DataLoader:
        """Train dataloader."""
        return data.ReaxDataLoader(RandomDataset(32, 64))

    def val_dataloader(self) -> data.DataLoader:
        """Val dataloader."""
        return data.ReaxDataLoader(RandomDataset(32, 64))

    def test_dataloader(self) -> data.DataLoader:
        """Test dataloader."""
        return data.ReaxDataLoader(RandomDataset(32, 64))

    def predict_dataloader(self) -> data.DataLoader:
        """Predict dataloader."""
        return data.ReaxDataLoader(RandomDataset(32, 64))


class RandomDataset(collections.abc.Sequence):
    """
    Warning:
        This is meant for testing/debugging and is experimental.
    """

    def __init__(self, size: int, length: int):
        self.len = length
        self.data = np.random.normal(size=(length, size))

    def __getitem__(self, index: int) -> jax.typing.ArrayLike:
        """Getitem function."""
        return self.data[index]

    def __len__(self) -> int:
        """Len function."""
        return self.len


class RandomIterableDataset(Iterable[np.ndarray]):
    """
    Warning:
        This is meant for testing/debugging and is experimental.
    """

    def __init__(self, size: int, count: int):
        self.size = size
        self.count = count

    def __iter__(self) -> Iterator[np.ndarray]:
        for _ in range(self.count):
            yield np.random.normal(size=self.size)


class BoringDataModule(reax.DataModule):
    """
    Warning:
        This is meant for testing/debugging and is experimental.
    """

    def __init__(self) -> None:
        super().__init__()
        self.random_full = RandomDataset(32, 64 * 4)

    def setup(self, stage: reax.Stage, /) -> None:
        # pylint: disable=attribute-defined-outside-init
        if stage.name == "fit":
            self.random_train = reax.data.Subset(self.random_full, indices=range(64))

        if stage.name in ("fit", "validate"):
            self.random_val = reax.data.Subset(self.random_full, indices=range(64, 64 * 2))

        if stage.name == "test":
            self.random_test = reax.data.Subset(self.random_full, indices=range(64 * 2, 64 * 3))

        if stage.name == "predict":
            self.random_predict = reax.data.Subset(self.random_full, indices=range(64 * 3, 64 * 4))

    def train_dataloader(self) -> reax.DataLoader:
        return reax.ReaxDataLoader(self.random_train)

    def val_dataloader(self) -> reax.DataLoader:
        return reax.ReaxDataLoader(self.random_val)

    def test_dataloader(self) -> reax.DataLoader:
        return reax.ReaxDataLoader(self.random_test)

    def predict_dataloader(self) -> reax.DataLoader:
        return reax.ReaxDataLoader(self.random_predict)



from collections.abc import Callable, Iterable
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypedDict,
    TypeVar,
    cast,
)

import beartype
import jax
import jaxtyping as jt

from .. import data
from ..utils import events

if TYPE_CHECKING:
    import reax

__all__ = ("MetricResults", "StageListener")

_T_co = TypeVar("_T_co", covariant=True)
_U = TypeVar("_U")
StageEvents = events.EventGenerator["StageListener"]


class StageListener:
    def on_stage_start(self, stage: "reax.Stage", /):
        """The stage is starting."""

    def on_stage_started(self, stage: "reax.Stage", /):
        """The stage is starting."""

    def on_stage_iter_starting(self, stage: "reax.Stage", step: int, /):
        """The stage is about to start an iteration."""

    def on_stage_iter_ending(self, stage: "reax.Stage", step: int, outputs: Any, /):
        """The stage just finished processing an iteration."""

    def on_stage_iter_ended(self, stage: "reax.Stage", step: int, outputs: Any, /):
        """The stage just finished processing an iteration."""

    def on_stage_end(self, stage: "reax.Stage", /):
        """The stage is ending."""

    def on_stage_ended(self, stage: "reax.Stage", /):
        """The stage has ended."""


class MetricResults(TypedDict):
    callback: dict[str, jax.Array]
    log: dict[str, jax.Array]
    pbar: dict[str, float]


class Stopper:
    """Class used by loops to manage their stop conditions."""

    def __init__(self):
        """Init function."""
        self._stop_requested = False
        self._conditions: list[Callable[[], bool]] = []

    @property
    def stop_requested(self) -> bool:
        """Stop requested."""
        return self._stop_requested

    @property
    def can_stop(self):
        """Can stop."""
        return all(condition() for condition in self._conditions)

    def do_stop(self) -> bool:
        """Do stop."""
        return self._stop_requested and self.can_stop

    def set(self):
        """Set function."""
        self._stop_requested = True

    def get(self) -> bool:
        """Get function."""
        return self._stop_requested

    def add_condition(self, condition: Callable[[], bool]) -> None:
        """Add condition."""
        self._conditions.append(condition)


@jt.jaxtyped(typechecker=beartype.beartype)
def batches_limit(
    batch_limit: int | float | None, dataloader: "reax.DataLoader"
) -> int | float | None:
    """Return a maximum number of batches given a dataloader and an optional batches limit.

    If the dataloader has fewer entries than the batch limit, then this will be used, otherwise
    the batches limit.

    Note:
        Will return `None` if there is no limit.

    Args:
        batch_limit (int | float): The batches limit.
        dataloader ("reax.DataLoader"): The dataloader.

    Returns:
        The maximum number of batches.
    """
    dataloader_size = data.sized_len(dataloader)
    if isinstance(batch_limit, int):
        if dataloader_size is None:
            return batch_limit

        return min(batch_limit, dataloader_size)

    if isinstance(batch_limit, float):
        if batch_limit in (None, 1.0):
            if dataloader_size is not None:
                return dataloader_size
            return None

        if dataloader_size is not None:
            # batch_limit is a finite float and we have a dataloader size
            batch_limit = cast(float, batch_limit)
            return int(round(batch_limit * dataloader_size))

        raise ValueError(
            f"Cannot determine number of batches from dataloader and batch_limit is "
            f"{batch_limit}"
        )

    # We can't say anything other than just 'go to the end'
    return float("inf")


class DataSourceManager(Generic[_T_co]):
    class LoaderProxy(data.DataLoader[_T_co, _U]):
        def __init__(self, manager: "DataSourceManager[_T_co]", method_name: str):
            self._manager = manager
            self._method_name = method_name
            self._iterable: Iterable[_U] | None = None

        def __iter__(self):
            return iter(self.dataloader)

        def __len__(self):
            return len(self.dataloader)

        def __getitem__(self, index):
            return self.dataloader.__getitem__(index)

        @property
        def sampler(self) -> "reax.data.Sampler":
            return self.dataloader.sampler

        @property
        def dataset(self) -> "reax.data.Dataset[_T_co]":
            return self.dataloader.dataset

        @property
        def dataloader(self) -> "reax.data.DataLoader[_T_co, _U]":
            if self._iterable is None:
                self._iterable = getattr(self._manager.source, self._method_name)()
            return self._iterable

        def with_new_sampler(
            self, sampler: "reax.data.Sampler"
        ) -> "reax.data.DataLoader[_T_co, _U]":
            return self.dataloader.with_new_sampler(sampler)

    def __init__(self, source: "reax.data.DataSource[_T_co]"):
        self._source = source
        self._ready = False

    @property
    def ready(self) -> bool:
        return self._ready

    def get_loader_proxy(self, method_name: str) -> LoaderProxy[_T_co, _U]:
        return DataSourceManager.LoaderProxy(self, method_name)

    @property
    def source(self) -> "reax.data.DataSource[_T_co]":
        if not self._ready:
            raise RuntimeError(
                "`prepare_and_setup` has not been called, this must be done before accessing the "
                "dataset"
            )
        return self._source

    def prepare_and_setup(self, stage) -> None:
        if self.ready:
            # Already done
            return

        self._source.prepare_data()
        self._source.setup(stage)
        self._ready = True


def get_datasource(
    datamodule: "reax.DataModule[_T_co] | None" = None,
    module: "reax.Module | None" = None,
) -> DataSourceManager[_T_co] | None:
    if datamodule is not None:
        return DataSourceManager(datamodule)

    if module is not None:
        return DataSourceManager(module)

    return None



from collections.abc import Callable
import contextlib
from typing import Generic, TypeVar
import uuid

ListenerT = TypeVar("ListenerT")

HANDLE = uuid.UUID


class EventGenerator(Generic[ListenerT]):
    """Manage listeners and fire events."""

    def __init__(self, default_args=tuple()):
        self._default_args = default_args
        self._event_listeners: dict[HANDLE, ListenerT] = {}

    def add_listener(self, listener: ListenerT) -> uuid.UUID:
        """Add listener."""
        handle = uuid.uuid4()
        self._event_listeners[handle] = listener
        return handle

    def remove_listener(self, handle: HANDLE) -> ListenerT:
        """Remove listener."""
        return self._event_listeners.pop(handle)

    def fire_event(self, event_fn: Callable, *args, **kwargs):
        """Fire event."""
        args = self._default_args + args
        for listener in list(self._event_listeners.values()):
            getattr(listener, event_fn.__name__)(*args, **kwargs)

    @contextlib.contextmanager
    def listen_context(self, *listener: ListenerT):
        """Listen context."""
        uuids = tuple()
        try:
            uuids = tuple(map(self.add_listener, listener))
            yield
        finally:
            tuple(map(self.remove_listener, uuids))

    T = TypeVar("T", bound=ListenerT)

    def get(self, handle: HANDLE) -> T:
        """Get a listener using its handle."""
        return self._event_listeners[handle]

    def find(
        # pylint: disable=redefined-builtin
        self,
        *,
        type: type[T] = None,
    ) -> list[T]:
        """Find listeners matching the passed filter(s)."""

        def filtr(listener: ListenerT) -> bool:
            if type is not None and not isinstance(listener, type):
                return False

            return True

        return [listener for listener in self._event_listeners.values() if filtr(listener)]
